#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Test the Core CheckAndGetProdProxy."""

from __future__ import absolute_import
import unittest
from mock import MagicMock as Mock, patch
from DIRAC import S_OK, S_ERROR
from ILCDIRAC.Core.Utilities.CheckAndGetProdProxy import checkAndGetProdProxy, checkOrGetGroupProxy

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.Core.Utilities.CheckAndGetProdProxy'


class CheckProxyTest(unittest.TestCase):
  """Test the CheckProxy."""

  def setUp(self):
    pass

  def tearDown(self):
    pass

  def test_success(self):
    """test for CheckandGetProdProxy: success......................................................."""
    with patch("%s.call" % MODULE_NAME, new=Mock(return_value=0)), \
         patch("%s.getProxyInfo" % MODULE_NAME, new=Mock(return_value=S_OK({"group": "ilc_prod"}))):
      res = checkAndGetProdProxy()
      self.assertTrue(res['OK'])

  def test_success_2(self):
    """test for CheckandGetProdProxy: sucess 2......................................................"""
    with patch("%s.call" % MODULE_NAME, new=Mock(return_value=0)), \
         patch("%s.getProxyInfo" % MODULE_NAME, new=Mock(side_effect=[S_OK({"group": "ilc_user"}), S_OK({"group": "ilc_prod"})])):
      res = checkAndGetProdProxy()
      self.assertTrue(res['OK'])

  def test_failure(self):
    """test for CheckandGetProdProxy: failure......................................................."""
    with patch("%s.call" % MODULE_NAME, new=Mock(return_value=1)), \
         patch("%s.getProxyInfo" % MODULE_NAME, new=Mock(return_value=S_ERROR('No proxy info'))):
      res = checkAndGetProdProxy()
      self.assertFalse(res['OK'])

  def test_failure_2(self):
    """test for CheckandGetProdProxy: semi failure 1................................................"""
    with patch("%s.call" % MODULE_NAME, new=Mock(return_value=0)), \
         patch("%s.getProxyInfo" % MODULE_NAME, new=Mock(return_value=S_ERROR('message'))):
      res = checkAndGetProdProxy()
      self.assertFalse(res['OK'])

  def test_failure_3(self):
    """test for CheckandGetProdProxy: semi failure 2................................................"""
    with patch("%s.call" % MODULE_NAME, new=Mock(return_value=0)), \
         patch("%s.getProxyInfo" % MODULE_NAME, new=Mock(side_effect=[S_OK({}), S_OK({"group": "ilc_prod"})])):
      res = checkAndGetProdProxy()
      self.assertTrue(res['OK'])

  def test_failure_4(self):
    """test for CheckandGetProdProxy: semi failure 3................................................"""
    with patch("%s.call" % MODULE_NAME, new=Mock(return_value=0)), \
         patch("%s.getProxyInfo" % MODULE_NAME, new=Mock(side_effect=[S_ERROR("no proxy info"), S_OK({"group": "ilc_prod"})])):
      res = checkAndGetProdProxy()
      self.assertTrue(res['OK'])

  def test_failure_5(self):
    """test for CheckandGetProdProxy: semi failure 4................................................"""
    with patch("%s.call" % MODULE_NAME, new=Mock(return_value=0)), \
         patch("%s.getProxyInfo" % MODULE_NAME, new=Mock(side_effect=[S_ERROR("no proxy info"), S_OK({"notgroup": "ilc_user"})])):
      res = checkAndGetProdProxy()
      self.assertTrue(not res['OK'])

  def test_failure_6(self):
    """test for CheckandGetProdProxy: semi failure 5................................................"""
    with patch("%s.call" % MODULE_NAME, new=Mock(return_value=0)), \
         patch("%s.getProxyInfo" % MODULE_NAME, new=Mock(side_effect=[S_ERROR("no proxy info"), S_OK({"group": "ilc_user"})])):
      res = checkAndGetProdProxy()
      self.assertTrue(not res['OK'])

  def test_failure_7(self):
    """test for CheckandGetProdProxy: semi failure 6................................................"""
    with patch("%s.call" % MODULE_NAME, new=Mock(return_value=0)), \
         patch("%s.getProxyInfo" % MODULE_NAME, new=Mock(side_effect=[S_ERROR("no proxy info"), S_ERROR("Still no proxy")])):
      res = checkAndGetProdProxy()
      self.assertTrue(not res['OK'])

  def test_checkOrGet(self):
    """test for checkOrGetGroupProxy: .............................................................."""
    with patch('%s.call' % MODULE_NAME, new=Mock(return_value=0)), \
         patch('%s.getProxyInfo' % MODULE_NAME, new=Mock(return_value=S_OK({'group': 'fcc_prod'}))):
      res = checkOrGetGroupProxy(['ilc_prod', 'fcc_prod'])
      self.assertTrue(res['OK'])
      self.assertEqual(res['Value'], 'fcc_prod')

    with patch('%s.call' % MODULE_NAME, new=Mock(return_value=0)), \
         patch('%s.getProxyInfo' % MODULE_NAME, new=Mock(return_value=S_OK({'groupProperties': 'NormalUser'}))):
      res = checkOrGetGroupProxy(['ilc_prod', 'fcc_prod'])
      self.assertFalse(res['OK'])
      self.assertIn('More than one', res['Message'])

    with patch('%s.call' % MODULE_NAME, new=Mock(return_value=0)), \
         patch('%s.getProxyInfo' % MODULE_NAME, new=Mock(return_value=S_OK({'group': 'ilc_user'}))):
      res = checkOrGetGroupProxy('ilc_user')
      self.assertTrue(res['OK'])
      self.assertEqual(res['Value'], 'ilc_user')


if __name__ == "__main__":
  SUITE = unittest.defaultTestLoader.loadTestsFromTestCase(CheckProxyTest)
  TESTRESULT = unittest.TextTestRunner(verbosity=2).run(SUITE)



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""
Marlin: Reconstructor after Mokka
"""
from __future__ import absolute_import
import types
import os

from ILCDIRAC.Interfaces.API.NewInterface.LCApplication import LCApplication
from ILCDIRAC.Core.Utilities.CheckXMLValidity import checkXMLValidity
from ILCDIRAC.Interfaces.Utilities.DDInterfaceMixin import DDInterfaceMixin

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Workflow.Parameter import Parameter
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
import six

LOG = gLogger.getSubLogger(__name__)
__RCSID__ = "$Id$"


class Marlin(DDInterfaceMixin, LCApplication):
  """Call Marlin reconstructor (after Mokka simulator)

  Usage:

  >>> marlin = Marlin()
  >>> marlin.getInputFromApp(mo)
  >>> marlin.setSteeringfile('SteeringFile.xml')
  >>> marlin.setOutputRecFile('MyOutputRecFile.rec')
  >>> marlin.setOutputDstFile('MyOutputDstFile.dst')

  Use :func:`setExtraCLIArguments` if you want to add command line parameters
  needed for easy parameter scans and passing non-standard strings (like cuts)

  >>> marlin = Marlin()
  >>> ...
  >>> marlin.setExtraCLIArguments( "--myProcessor.myParameter=someValue" )
  >>> ...

  The output file for marlin is changed automatically if the xml
  steering file contains a processor called *MyLCIOOutputProcessor*

  >>> marlin = Marlin()
  >>> ...
  >>> marlin.setOutputFile( "output_job123.slcio" )
  >>> ...

  .. versionchanged:: v26r0p4

    The default value for the GearFile was removed. It used to be
    "GearOutput.xml" but as reconstruction based on DD4hep is not using gear
    files a default value is more dangerous than before.

  .. versionchanged:: v29r0p4

    New member flag ``keepRecFile``, defaults to ``True``, if set to
    ``False``, the REC file is not uploaded from ProductionJobs
  """

  def __init__(self, paramdict=None):

    self.outputDstPath = ''
    self.outputDstFile = ''
    self.outputRecPath = ''
    self.outputRecFile = ''
    self.keepRecFile = True
    self.gearFile = ''
    self.processorsToUse = []
    self.processorsToExclude = []
    super(Marlin, self).__init__(paramdict)
    # Those 5 need to come after default constructor
    self._modulename = 'MarlinAnalysis'
    self._moduledescription = 'Module to run MARLIN'
    self.appname = 'marlin'
    self.datatype = 'REC'
    self.detectortype = 'ILD'
    self.detectorModel = ''
    self._ops = Operations()

  def setGearFile(self, gearFile):
    """Define input gear file for Marlin.

    :param str gearFile: input gear file for Marlin reconstructor
    """
    self._checkArgs({'gearFile': (str,)})

    self.gearFile = gearFile
    if os.path.exists(gearFile) or gearFile.lower().count("lfn:"):
      self.inputSB.append(gearFile)

  def setKeepRecFile(self, val):
    """Set the ``keepRecFile`` flag.

    Only relevant for ProductionJobs

    :param bool val: If ``False`` REC file is not stored
    """
    self._checkArgs({'val': bool})
    self.keepRecFile = val

  def setOutputRecFile(self, outputRecFile, path=None):
    """Optional: Define output rec file for Marlin. Used only in production
    context. Use :func:`UserJob.setOutputData
    <ILCDIRAC.Interfaces.API.NewInterface.UserJob.UserJob.setOutputData>` if you
    want to keep the file on the grid.

    :param str outputRecFile: output rec file for Marlin
    :param str path: Path where to store the file.

    """
    self._checkArgs({'outputRecFile': (str,)})
    self.outputRecFile = outputRecFile
    self.prodparameters[self.outputRecFile] = {}
    self.prodparameters[self.outputRecFile]['datatype'] = 'REC'
    if path:
      self.outputRecPath = path

  def setOutputDstFile(self, outputDstFile, path=None):
    """Optional: Define output dst file for Marlin.  Used only in production
    context. Use :func:`UserJob.setOutputData
    <ILCDIRAC.Interfaces.API.NewInterface.UserJob.UserJob.setOutputData>` if you
    want to keep the file on the grid.

    :param str outputDstFile: output dst file for Marlin
    :param str path: Path where to store the file.

    """
    self._checkArgs({'outputDstFile': (str,)})
    self.outputDstFile = outputDstFile
    self.prodparameters[self.outputDstFile] = {}
    self.prodparameters[self.outputDstFile]['datatype'] = 'DST'
    if path:
      self.outputDstPath = path

  def setProcessorsToUse(self, processorlist):
    """Define processor list to use.

    Overwrite the default list (full reco). Useful for users willing to do dedicated analysis (TPC, Vertex digi, etc.)

    >>> ma.setProcessorsToUse(['libMarlinTPC.so','libMarlinReco.so','libOverlay.so','libMarlinTrkProcessors.so'])

    :param processorlist: list of processors to use
    :type processorlist: list
    """
    self._checkArgs({'processorlist': list})
    self.processorsToUse = processorlist

  def setProcessorsToExclude(self, processorlist):
    """Define processor list to exclude.

    Overwrite the default list (full reco). Useful for users willing to do dedicated analysis (TPC, Vertex digi, etc.)

    >>> ma.setProcessorsToExclude(['libLCFIVertex.so'])

    :param processorlist: list of processors to exclude
    :type processorlist: list
    """
    self._checkArgs({'processorlist': list})
    self.processorsToExclude = processorlist

  def _userjobmodules(self, stepdefinition):
    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    res2 = self._setUserJobFinalization(stepdefinition)
    if not res1["OK"] or not res2["OK"]:
      return S_ERROR('userjobmodules failed')
    return S_OK()

  def _prodjobmodules(self, stepdefinition):

    # Here one needs to take care of listoutput
    if self.outputPath:
      self._listofoutput.append({'OutputFile': '@{OutputFile}', "outputPath": "@{OutputPath}",
                                 "outputDataSE": '@{OutputSE}'})

    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    res2 = self._setOutputComputeDataList(stepdefinition)
    if not res1["OK"] or not res2["OK"]:
      return S_ERROR('prodjobmodules failed')
    return S_OK()

  def _checkConsistency(self, job=None):

    if not self.version:
      return S_ERROR('Version not set!')

    if self.steeringFile:
      # FIXME: delete dead code
      # if not os.path.exists(self.steeringFile) and not self.steeringFile.lower().count("lfn:"):
        ##res = Exists(self.SteeringFile)
        #res = S_OK()
        # if not res['OK']:
        # return res
      if os.path.exists(self.steeringFile):
        res = checkXMLValidity(self.steeringFile)
        if not res['OK']:
          return S_ERROR("Supplied steering file cannot be read with xml parser: %s" % (res['Message']))

    if not self.gearFile:
      LOG.info('GEAR file not given, will not use any gear file')
    # FIXME: delete dead code
    # if self.gearFile:
      # if not os.path.exists(self.gearFile) and not self.gearFile.lower().count("lfn:"):
      ##res = Exists(self.gearFile)
      #res = S_OK()
      # if not res['OK']:
      # return res

    if self._jobtype != 'User':
      if not self.outputFile:
        self._listofoutput.append({'outputFile': '@{outputDST}', 'outputPath': '@{outputPathDST}',
                                   'outputDataSE': '@{OutputSE}'})
        if self.keepRecFile:
          self._listofoutput.append({'outputFile': '@{outputREC}', 'outputPath': '@{outputPathREC}',
                                     'outputDataSE': '@{OutputSE}'})
      self.prodparameters['detectorType'] = self.detectortype
      self.prodparameters['marlin_gearfile'] = self.gearFile
      self.prodparameters['marlin_steeringfile'] = self.steeringFile

    return S_OK()

  def _applicationModule(self):

    md1 = self._createModuleDefinition()
    md1.addParameter(Parameter("inputGEAR", '', "string", "", "", False, False,
                               "Input GEAR file"))
    md1.addParameter(Parameter("detectorModel", '', "string", "", "", False, False,
                               "DD4hep Geomtry File"))
    md1.addParameter(Parameter("ProcessorListToUse", [], "list", "", "", False, False,
                               "List of processors to use"))
    md1.addParameter(Parameter("ProcessorListToExclude", [], "list", "", "", False, False,
                               "List of processors to exclude"))
    md1.addParameter(Parameter("debug", False, "bool", "", "", False, False,
                               "debug mode"))
    return md1

  def _applicationModuleValues(self, moduleinstance):

    moduleinstance.setValue("inputGEAR", self.gearFile)
    moduleinstance.setValue("detectorModel", self.detectorModel)
    moduleinstance.setValue('ProcessorListToUse', self.processorsToUse)
    moduleinstance.setValue('ProcessorListToExclude', self.processorsToExclude)
    moduleinstance.setValue("debug", self.debug)

  def _checkWorkflowConsistency(self):
    return self._checkRequiredApp()

  def _resolveLinkedStepParameters(self, stepinstance):
    if isinstance(self._linkedidx, six.integer_types):
      self._inputappstep = self._jobsteps[self._linkedidx]
    if self._inputappstep:
      stepinstance.setLink("InputFile", self._inputappstep.getType(), "OutputFile")
    return S_OK()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Installs properly ILD soft and SiD soft, and all dependencies.

:since: Feb 4, 2010
:author: Stephane Poss
:author: Przemyslaw Majewski
"""
from __future__ import absolute_import
import os
import tarfile
import zipfile
import DIRAC

from DIRAC import S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.Utilities.Subprocess import systemCall

from ILCDIRAC.Core.Utilities.DetectOS import NativeMachine
from ILCDIRAC.Core.Utilities.ResolveDependencies import resolveDeps
from ILCDIRAC.Core.Utilities.TARsoft import installInAnyArea, installDependencies
import six

LOG = DIRAC.gLogger.getSubLogger(__name__)

__RCSID__ = "$Id$"

natOS = NativeMachine()


class CombinedSoftwareInstallation(object):
  """Combined means that it will try to install in the Shared area and in the LocalArea, depending on the user's rights."""

  def __init__(self, argumentsDict):
    """Standard constructor.

    Defines, from dictionary of job parameters passed, a set of members to hold e.g. the
    applications and the system config.

    Also determines the SharedArea and LocalArea.
    """

    self.ops = Operations()

    self.job = argumentsDict.get('Job', {})
    self.ce = argumentsDict.get('CE', {})
    self.source = argumentsDict.get('Source', {})

    apps = []
    if 'SoftwarePackages' in self.job:
      if isinstance(self.job['SoftwarePackages'], six.string_types):
        apps = [self.job['SoftwarePackages']]
      elif isinstance(self.job['SoftwarePackages'], list):
        apps = self.job['SoftwarePackages']

    self.jobConfig = ''
    if 'SystemConfig' in self.job:
      self.jobConfig = self.job['SystemConfig']
    elif 'Platform' in self.job:
      self.jobConfig = self.job['Platform']
    else:
      self.jobConfig = natOS.CMTSupportedConfig()[0]

    self.apps = []
    for app in apps:
      LOG.verbose('Requested Package %s' % app)
      app = tuple(app.split('.'))
      if len(app) > 2:
        tempapp = app
        app = []
        app.append(tempapp[0])
        app.append(".".join(tempapp[1:]))

      deps = resolveDeps(self.jobConfig, app[0], app[1])
      for dep in deps:
        depapp = (dep['app'], dep['version'])
        self.apps.append(depapp)
      self.apps.append(app)

    self.ceConfigs = []
    if 'CompatiblePlatforms' in self.ce:
      self.ceConfigs = self.ce['CompatiblePlatforms']
      if isinstance(self.ceConfigs, six.string_types):
        self.ceConfigs = [self.ceConfigs]
    # else:
    # Use always the list of compatible platform.
    self.ceConfigs = natOS.CMTSupportedConfig()

    self.sharedArea = getSharedAreaLocation()
    LOG.info("SharedArea is %s" % self.sharedArea)
    self.localArea = getLocalAreaLocation()
    LOG.info("LocalArea is %s" % self.localArea)

  def execute(self):
    """Main method of the class executed by DIRAC jobAgent.

    Executes the following:
      - look for the compatible platforms in the CS, see if one matches request
      - install the applications, calls :mod:`~ILCDIRAC.Core.Utilities.TARsoft`

    :return: S_OK(), S_ERROR()
    """
    if not self.apps:
      # There is nothing to do
      return DIRAC.S_OK()
    if not self.jobConfig:
      LOG.error('No architecture requested')
      return DIRAC.S_ERROR('No architecture requested')

    found_config = False

    LOG.info("Found CE Configs %s, compatible with system reqs %s" % (",".join(self.ceConfigs),
                                                                                self.jobConfig))
    res = self.ops.getSections('/AvailableTarBalls')
    if not res['OK']:
      return res
    else:
      supported_systems = res['Value']
      # look in CS if specified platform has software available. Normally consistency check is done at submission time
      for ceConfig in self.ceConfigs:
        for supp_systems in supported_systems:
          if ceConfig == supp_systems:
            self.jobConfig = ceConfig
            found_config = True
            break
    if not found_config:
      if self.ceConfigs:  # redundant check as this is done in the job agent, if locally running option might not be defined
        LOG.error('Requested architecture not supported by CE')
        return DIRAC.S_ERROR('Requested architecture not supported by CE')
      else:
        LOG.info('Assume locally running job, will install software in ')

    areas = []
    # Deal with shared/local area: first try to see if the Shared area exists
    # and if not create it (try to). If it fails, fall back to local area
    if not self.sharedArea:
      if createSharedArea():
        self.sharedArea = getSharedAreaLocation()
        if self.sharedArea:
          areas.append(self.sharedArea)
          LOG.info("Will attempt to install in shared area")
    else:
      areas.append(self.sharedArea)
    areas.append(self.localArea)

    for app in self.apps:
      res = checkCVMFS(self.jobConfig, app)
      if res['OK']:
        LOG.notice('Software %s is available on CVMFS, skipping' % ", ".join(app))
        continue

      # First install the original package in any of the areas
      resInstall = installInAnyArea(areas, app, self.jobConfig)
      if not resInstall['OK']:
        return resInstall

      # Then install the dependencies, which can be in a different area
      resDeps = installDependencies(app, self.jobConfig, areas)
      if not resDeps['OK']:
        LOG.error("Failed to install dependencies: ", resDeps['Message'])
        return S_ERROR("Failed to install dependencies")

    if self.sharedArea:
      # List content
      listAreaDirectory(self.sharedArea)

    return DIRAC.S_OK()


def listAreaDirectory(area):
  """List the content of the given area."""
  LOG.info("Listing content of area %s :" % (area))
  res = systemCall(5, ['ls', '-al', area])
  if not res['OK']:
    LOG.error('Failed to list the area directory', res['Message'])
  elif res['Value'][0]:
    LOG.error('Failed to list the area directory', res['Value'][2])
  else:
    # no timeout and exit code is 0
    LOG.info(res['Value'][1])


def log(n, line):
  """print line."""
  LOG.info(line)


def getSharedAreaLocation():
  """Discover location of Shared SW area based on the list of "Operations/Software/SharedAreaLocations".

  Environment variables are allowed as well. First match is returned. Default list is: "/cvmfs/ilc.desy.de/clic",
  "$VO_ILC_SW_DIR", "$OSG_APP", "/cvmfs/oasis.opensciencegrid.org/ilc/clic".

  :returns: path to shared area
  :rtype: str
  """

  listOfSharedAreas = Operations().getValue("Software/SharedAreaLocations",
                                             ["/cvmfs/ilc.desy.de/clic",
                                               "$VO_ILC_SW_DIR",
                                               "$OSG_APP",
                                               "/cvmfs/oasis.opensciencegrid.org/ilc/clic",
                                             ])
  if not isinstance(listOfSharedAreas, list):
    listOfSharedAreas = [listOfSharedAreas]
  LOG.debug("ListOfSharedAreas: %s " % ", ".join(listOfSharedAreas))
  sharedArea = ''
  for area in listOfSharedAreas:
    LOG.debug("Checking SharedArea %s " % area)
    if area.startswith("$"):  # is an environment variable
      envVar = area[1:]
      if envVar in os.environ:
        sharedArea = os.path.join(os.environ[envVar], 'clic')
    else:
      sharedArea = area

    if os.path.exists(sharedArea):
      LOG.info("Using shared area %s based on %s " % (sharedArea, area))
      break

  if DIRAC.gConfig.getValue('/LocalSite/SharedArea', ''):
    sharedArea = DIRAC.gConfig.getValue('/LocalSite/SharedArea')
    LOG.debug('Using CE SharedArea at "%s"' % sharedArea)

  if len(sharedArea):
    # if defined, check that it really exists
    if not os.path.isdir(sharedArea):
      LOG.warn('Missing Shared Area Directory:', sharedArea)
      sharedArea = ''

  return sharedArea


def createSharedArea():
  """Method to be used by SAM jobs to make sure the proper directory structure is created if it does not exists."""
  if 'VO_ILC_SW_DIR' not in os.environ and "OSG_APP" not in os.environ:
    LOG.info('VO_ILC_SW_DIR and OSG_APP not defined.')
    return False
  sharedArea = '.'
  if 'VO_ILC_SW_DIR' in os.environ:
    sharedArea = os.environ['VO_ILC_SW_DIR']
  elif "OSG_APP" in os.environ:
    sharedArea = os.environ["OSG_APP"]
  if sharedArea == '.':
    LOG.info('VO_ILC_SW_DIR or OSG_APP points to "."')
    return False

  # if not os.path.isdir( sharedArea ):
  #  LOG.error( 'VO_ILC_SW_DIR="%s" is not a directory' % sharedArea )
  #  return False

  sharedArea = os.path.join(sharedArea, 'clic')
  try:
    if os.path.isdir(sharedArea) and not os.path.islink(sharedArea):
      return True
    if not os.path.exists(sharedArea):
      os.makedirs(sharedArea)
      return True
    os.remove(sharedArea)
    os.makedirs(sharedArea)
    return True
  except OSError as x:
    LOG.error('Problem trying to create shared area', str(x))
    return False


def getLocalAreaLocation():
  """Discover Location of Local SW Area.

  This area is populated by DIRAC job Agent for jobs needing SW not present
  in the Shared Area.
  Currently is always the location used as the software is not in the shared area.
  """
  if DIRAC.gConfig.getValue('/LocalSite/LocalArea', ''):
    localArea = DIRAC.gConfig.getValue('/LocalSite/LocalArea')
  else:
    localArea = os.path.join(DIRAC.rootPath, 'LocalArea')

  # check if already existing directory
  if not os.path.isdir(localArea):
    # check if we can create it
    if os.path.exists(localArea):
      try:
        os.remove(localArea)
      except OSError as err:
        LOG.error('Cannot remove:', localArea + " because " + str(err))
        localArea = ''
    else:
      try:
        os.mkdir(localArea)
      except OSError as err:
        LOG.error('Cannot create:', localArea + " because " + str(err))
        localArea = ''
  return localArea


def getSoftwareFolder(platform, appname, appversion):
  """Discover location of a given folder, either the local or the shared area.

  :param str platform: platform
  :param str appname: name of the application
  :param str appversion: version of the application
  """
  res = checkCVMFS(platform, [appname, appversion])
  if res["OK"]:
    return S_OK(res['Value'][0])

  app_tar = Operations().getValue('/AvailableTarBalls/%s/%s/%s/TarBall' % (platform, appname, appversion), '')
  if not app_tar:
    return S_ERROR("Could not find %s, %s name from CS" % (appname, appversion))
  if app_tar.count("gz"):
    folder = app_tar.replace(".tgz", "").replace(".tar.gz", "")
  else:
    folder = app_tar

  localArea = getLocalAreaLocation()
  sharedArea = getSharedAreaLocation()
  if os.path.exists(os.path.join(localArea, folder)):
    mySoftwareRoot = localArea
  elif os.path.exists(os.path.join(sharedArea, folder)):
    mySoftwareRoot = sharedArea
  else:
    return S_ERROR('Missing installation of %s!' % folder)
  mySoftDir = os.path.join(mySoftwareRoot, folder)
  return S_OK(mySoftDir)


def getEnvironmentScript(platform, appname, appversion, fcn_env):
  """Return the path to the environment script, either from CVMFS, or from the fcn_env function.

  :param str platform: platform
  :param str appname: name of the application
  :param str appversion: version of the application

  :param fcn_env: function provided by the Application daughter class
    to create its own environment in case there is no environment script in the
    application version. Signature has to be *fcn_env(platform, appname,
    appversion)* and has to return *S_OK(pathToScript)*. If you do not want to
    provide such a function you can set fcn_env=S_ERROR('User must provide script')
    and the error will be returned.
  :type fcn_env: ``function``
  """
  res = checkCVMFS(platform, [appname, appversion])
  if res["OK"]:
    cvmfsenv = res['Value'][1]
    LOG.info("SoftwareInstallation: CVMFS Script found: %s" % cvmfsenv)
    if cvmfsenv:
      return S_OK(cvmfsenv)

  # If user does not provide enviroment creation function return the dictionary
  if isinstance(fcn_env, dict):
    return fcn_env

  # if CVMFS script is not here, the module produces its own.
  return fcn_env(platform, appname, appversion)


def checkCVMFS(platform, app):
  """Check the existence of the CVMFS path for given platform and application.

  :param str platform: platform
  :param app: tuple off application name and version
  :type app: ``tuple``
  :returns: S_OK of tuple of path and environmen stript
  """
  name, version = app
  csPath = "/AvailableTarBalls/%s/%s/%s" % (platform, name, version)
  cvmfspath = Operations().getValue(csPath + "/CVMFSPath", "")
  envScript = Operations().getValue(csPath + "/CVMFSEnvScript", "")
  if cvmfspath and os.path.exists(cvmfspath):
    return S_OK((cvmfspath, envScript))
  message = "Package %s version %s not found natively on cvmfs," % (name, version)
  LOG.info(message + " will try to use shared area instead: " + str(Operations().getOptionsDict(csPath)))
  return S_ERROR('Missing CVMFS!')


def unzip_file_into_dir(myfile, mydir):
  """Used to unzip the downloaded detector model."""
  zfobj = zipfile.ZipFile(myfile)
  zfobj.extractall(path=mydir)


def extractTarball(tarball, directory):
  """Extract tarball into directory.

  :param str tarball: path to the tarball
  :param str directory: path to the directory
  :returns: ``S_OK``, ``S_ERROR``
  """
  LOG.info('Extracting files', 'from %r to %r' % (tarball, directory))
  try:
    if tarball.endswith(('.tar', 'tar.gz')):
      tarball = tarfile.open(tarball, mode='r:*')
    else:
      return S_ERROR('Unknown tarball format')
    tarball.extractall(directory)
  except (RuntimeError, OSError) as e:
    LOG.error('Failed to extract tarball', repr(e))
    return S_ERROR('Failed to extract tarball')
  return S_OK()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Run LCSIM.

Called by Job Agent.

:since: Apr 7, 2010

:author: Stephane Poss
"""

from __future__ import absolute_import
import os
import shutil

from DIRAC.Core.Utilities.Subprocess import shellCall
from DIRAC import S_OK, S_ERROR, gLogger

from ILCDIRAC.Workflow.Modules.ModuleBase import ModuleBase
from ILCDIRAC.Workflow.Utilities.CompactMixin import CompactMixin
from ILCDIRAC.Core.Utilities.CombinedSoftwareInstallation import getSoftwareFolder
from ILCDIRAC.Core.Utilities.PrepareOptionFiles import prepareLCSIMFile, getNewLDLibs
from ILCDIRAC.Core.Utilities.resolvePathsAndNames import resolveIFpaths, getProdFilename
from ILCDIRAC.Core.Utilities.PrepareLibs import removeLibc
from ILCDIRAC.Core.Utilities.FindSteeringFileDir import getSteeringFileDirName

__RCSID__ = '$Id$'
LOG = gLogger.getSubLogger(__name__)


class LCSIMAnalysis(CompactMixin, ModuleBase):
  """Define the LCSIM analysis part of the workflow."""

  def __init__(self):
    super(LCSIMAnalysis, self).__init__()
    self.enable = True
    self.STEP_NUMBER = ''
    self.result = S_ERROR()
    self.sourcedir = ''
    self.SteeringFile = ''
    self.InputFile = []
    self.InputData = []
    self.outputREC = ""
    self.outputDST = ""
    self.aliasproperties = ''
    self.applicationName = 'LCSIM'
    self.eventstring = ['']
    self.extraparams = ''
    self.OutputFile = ''  # Set in ModuleBase
    self.detectorModel = ''
    self.trackingstrategy = ''
    self.NumberOfEvents = -1

  def applicationSpecificInputs(self):
    """Resolve all input variables for the module here.

    :return: S_OK()
    """

    if 'inputXML' in self.step_commons:
      self.SteeringFile = self.step_commons['inputXML']

    # TODO: Next is necessary for old interface, should be removed when old prods are archived.
    if 'outputREC' in self.step_commons:
      self.outputREC = self.step_commons['outputREC']

    if 'outputDST' in self.step_commons:
      self.outputDST = self.step_commons['outputDST']

    if 'inputSlcio' in self.step_commons:
      inputf = self.step_commons["inputSlcio"]
      if not isinstance(inputf, list):
        inputf = inputf.split(";")
      self.InputFile = inputf

    # FIXME: hardcode default detector model add check for detectorModel when submitting lcsim jobs!
      # because currently no detectormodel is required!
    if "IS_PROD" in self.workflow_commons:
      if not self.detectorModel:
        self.detectorModel = "clic_sid_cdr.zip"

    if 'ExtraParams' in self.step_commons:
      self.extraparams = self.step_commons['ExtraParams']

    if 'IS_PROD' in self.workflow_commons:
      if self.workflow_commons["IS_PROD"]:
        # self.outputREC = getProdFilename(self.outputREC,int(self.workflow_commons["PRODUCTION_ID"]),
        #                                 int(self.workflow_commons["JOB_ID"]))
        # self.outputDST = getProdFilename(self.outputDST,int(self.workflow_commons["PRODUCTION_ID"]),
        #                                 int(self.workflow_commons["JOB_ID"]))
        # if 'MokkaOutput' in self.workflow_commons:
        #  self.InputFile = getProdFilename(self.workflow_commons["MokkaOutput"],int(self.workflow_commons["PRODUCTION_ID"]),
        #                                    int(self.workflow_commons["JOB_ID"]))
        if 'ProductionOutputData' in self.workflow_commons:
          outputlist = self.workflow_commons['ProductionOutputData'].split(";")
          for obj in outputlist:
            if obj.lower().count("_rec_"):
              self.outputREC = os.path.basename(obj)
            elif obj.lower().count("_dst_"):
              self.outputDST = os.path.basename(obj)
            elif obj.lower().count("_sim_"):
              self.InputFile = [os.path.basename(obj)]
        else:
          self.outputREC = getProdFilename(self.outputREC, int(self.workflow_commons["PRODUCTION_ID"]),
                                           int(self.workflow_commons["JOB_ID"]))
          self.outputDST = getProdFilename(self.outputDST, int(self.workflow_commons["PRODUCTION_ID"]),
                                           int(self.workflow_commons["JOB_ID"]))
          if 'SLICOutput' in self.workflow_commons:
            self.InputFile = [getProdFilename(self.workflow_commons["SLICOutput"],
                                              int(self.workflow_commons["PRODUCTION_ID"]),
                                              int(self.workflow_commons["JOB_ID"]))]

    if 'aliasproperties' in self.step_commons:
      self.aliasproperties = self.step_commons["aliasproperties"]

    if not len(self.InputFile) and len(self.InputData):
      for files in self.InputData:
        if files.lower().find(".slcio") > -1:
          self.InputFile.append(files)
    LOG.info("Input files to treat %s" % self.InputFile)
    return S_OK('Parameters resolved')

  def runIt(self):
    """Called by JobAgent.

    Execute the following:
      - prepend in the LD_LIBRARY_PATH any lib directory of any dependency (e.g. root)
      - prepare the list of files to run on
      - set the cacheDirectory and put in there the alias.properties
      - set the lcsim file using :any:`prepareLCSIMFile`
      - run java and catch the exit code

    :return: S_OK(), S_ERROR()
    """
    self.result = S_OK()
    if not self.platform:
      self.result = S_ERROR('No ILC platform selected')
    elif not self.applicationLog:
      self.result = S_ERROR('No Log file provided')
    if not self.result['OK']:
      LOG.error("Failed to resolve input parameters:", self.result["Message"])
      return self.result

    if not self.workflowStatus['OK'] or not self.stepStatus['OK']:
      LOG.verbose('Workflow status = %s, step status = %s' % (self.workflowStatus['OK'], self.stepStatus['OK']))
      return S_OK('LCSIM should not proceed as previous step did not end properly')

    res = getSoftwareFolder(self.platform, self.applicationName, self.applicationVersion)
    if not res['OK']:
      LOG.error('LCSIM was not found in either the local area or shared area:', res['Message'])
      return res
    lcsim_name = res['Value']
    # Need to fetch the new LD_LIBRARY_PATH
    new_ld_lib_path = getNewLDLibs(self.platform, self.applicationName, self.applicationVersion)

    runonslcio = []
    if len(self.InputFile):
      res = resolveIFpaths(self.InputFile)
      if not res['OK']:
        self.setApplicationStatus('LCSIM: missing input slcio file')
        return S_ERROR('Missing slcio file!')
      runonslcio = res['Value']
    # for inputfile in inputfilelist:
    #  LOG.verbose("Will try using %s"%(os.path.basename(inputfile)))
    #  runonslcio.append(os.path.join(os.getcwd(),os.path.basename(inputfile)))

    retMod = self.downloadDetectorZip()
    if not retMod:
      return retMod

    # Collect jar files to put in classspath
    jars = []
    if os.path.exists("lib"):
      for libs in os.listdir("lib"):
        if os.path.basename(libs).find(".jar") > 0:
          jars.append(os.path.abspath(os.path.join("lib", libs)))
      new_ld_lib_path = "./lib:%s" % new_ld_lib_path
      # Remove any libc remaining in .lib
      removeLibc("./lib")

    # Define cache directory as local folder
    aliasproperties = os.path.basename(self.aliasproperties)
    cachedir = os.getcwd()
    if not os.path.isdir(os.path.join(cachedir, ".lcsim")):
      try:
        os.mkdir(os.path.join(cachedir, ".lcsim"))
      except OSError as x:
        LOG.error("Could not create .lcsim folder !", str(x))
    if os.path.exists(os.path.join(cachedir, ".lcsim")):
      lcsimfolder = os.path.join(cachedir, ".lcsim")
      if os.path.exists(aliasproperties):
        LOG.verbose("Copy alias.properties file in %s" % (lcsimfolder))
        shutil.copy(aliasproperties, os.path.join(lcsimfolder, aliasproperties))
      if os.path.exists(os.path.basename(self.detectorModel)):
        try:
          os.mkdir(os.path.join(lcsimfolder, "detectors"))
        except OSError as x:
          LOG.error("Could not create detectors folder !", str(x))
        if os.path.exists(os.path.join(lcsimfolder, "detectors")):
          LOG.verbose("Copy detector model.zip into the .lcsim/detectors folder")
          shutil.copy(os.path.basename(self.detectorModel),
                      os.path.join(lcsimfolder, "detectors", os.path.basename(self.detectorModel)))

    paths = {}
    paths[os.path.basename(self.SteeringFile)] = os.path.basename(self.SteeringFile)
    paths[os.path.basename(self.trackingstrategy)] = os.path.basename(self.trackingstrategy)
    for myfile in paths:
      if len(myfile):
        #file = os.path.basename(file)
        if not os.path.exists(myfile):
          res = getSteeringFileDirName(self.platform, self.applicationName, self.applicationVersion)
          if not res['OK']:
            LOG.error('Failed finding the steering file folder:', res["Message"])
            return res
          steeringfiledirname = res['Value']
          if os.path.exists(os.path.join(steeringfiledirname, myfile)):
            paths[myfile] = os.path.join(steeringfiledirname, myfile)
        if not os.path.exists(paths[myfile]):
          LOG.error('Failed finding file', paths[myfile])
          return S_ERROR("Could not find file %s" % paths[myfile])
    self.SteeringFile = paths[os.path.basename(self.SteeringFile)]
    self.trackingstrategy = paths[os.path.basename(self.trackingstrategy)]

    lcsimfile = "job_%s.lcsim" % self.STEP_NUMBER
    res = prepareLCSIMFile(self.SteeringFile, lcsimfile, self.NumberOfEvents,
                           self.trackingstrategy, runonslcio, jars, cachedir,
                           self.OutputFile, self.outputREC, self.outputDST, self.debug)
    if not res['OK']:
      LOG.error("Could not treat input lcsim file because %s" % res['Message'])
      return S_ERROR("Error creating lcsim file")
    else:
      LOG.verbose("File job.lcsim created properly")
    self.eventstring = [res['Value']]

    scriptName = 'LCSIM_%s_Run_%s.sh' % (self.applicationVersion, self.STEP_NUMBER)
    if os.path.exists(scriptName):
      os.remove(scriptName)
    script = open(scriptName, 'w')
    script.write('#!/bin/sh \n')
    script.write('#####################################################################\n')
    script.write('# Dynamically generated script to run a production or analysis job. #\n')
    script.write('#####################################################################\n')
    if new_ld_lib_path:
      script.write("declare -x LD_LIBRARY_PATH=%s\n" % new_ld_lib_path)
    script.write("declare -x JAVALIBPATH=./\n")
    if os.path.exists("lib"):
      script.write("declare -x JAVALIBPATH=./lib\n")
    script.write('echo =========\n')
    script.write('echo java version :\n')
    script.write('java -version\n')
    script.write('env | sort >> localEnv.log\n')
    script.write('echo =========\n')
    comm = "java -Xmx1536m -Xms256m -server -Djava.library.path=$JAVALIBPATH -Dorg.lcsim.cacheDir=%s -jar %s %s %s %s\n" % (
        cachedir, lcsim_name, self.extraparams, lcsimfile, self.extraCLIarguments)
    LOG.info("Will run %s" % comm)
    script.write(comm)
    script.write('declare -x appstatus=$?\n')
    script.write('exit $appstatus\n')
    script.close()
    if os.path.exists(self.applicationLog):
      os.remove(self.applicationLog)

    os.chmod(scriptName, 0o755)
    comm = 'sh -c "./%s"' % scriptName
    self.setApplicationStatus('LCSIM %s step %s' % (self.applicationVersion, self.STEP_NUMBER))
    self.stdError = ''
    self.result = shellCall(0, comm, callbackFunction=self.redirectLogOutput, bufferLimit=20971520)
    #self.result = {'OK':True,'Value':(0,'Disabled Execution','')}
    resultTuple = self.result['Value']
    if not os.path.exists(self.applicationLog):
      LOG.error("Something went terribly wrong, the log file is not present")
      self.setApplicationStatus('%s failed terribly, you are doomed!' % (self.applicationName))
      return S_ERROR('%s did not produce the expected log' % (self.applicationName))

    status = resultTuple[0]
    # stdOutput = resultTuple[1]
    # stdError = resultTuple[2]
    LOG.info("Status after the application execution is", str(status))

    return self.finalStatusReport(status)



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Created on Jul 25, 2011.

:author: Stephane Poss
"""

from __future__ import absolute_import
__RCSID__ = "$Id$"

from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC import S_OK, gLogger
from DIRAC.WorkloadManagementSystem.Client.JobMonitoringClient import JobMonitoringClient

from ILCDIRAC.OverlaySystem.Client.OverlaySystemClient import OverlaySystemClient

AGENT_NAME = 'Overlay/ResetCounters'


class ResetCounters (AgentModule):
  """Reset the number of jobs at all sites: some sites are not updated properly, so once in a while it's needed to restore the correct number of jobs.

  It does not need to be exact, but enough to clear some of the jobs.
  """

  def initialize(self):
    """Initialize the agent."""
    self.am_setOption("PollingTime", 60)
    self.ovc = OverlaySystemClient()
    self.jobmon = JobMonitoringClient()
    return S_OK()

  def execute(self):
    """This is called by the Agent Reactor."""
    res = self.ovc.getSites()
    if not res['OK']:
      return res
    sitedict = {}
    sites = res['Value']
    gLogger.info("Will update info for sites %s" % sites)
    for site in sites:
      attribdict = {"Site": site, "ApplicationStatus": 'Getting overlay files'}
      res = self.jobmon.getCurrentJobCounters(attribdict)
      if not res['OK']:
        continue
      if 'Running' in res['Value']:
        sitedict[site] = res['Value']['Running']
      else:
        sitedict[site] = 0
    gLogger.info("Setting new values %s" % sitedict)
    res = self.ovc.setJobsAtSites(sitedict)
    if not res['OK']:
      gLogger.error(res['Message'])
      return res

    return S_OK()



#!/usr/bin/env python
#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Test the RootExecutableAnalysis module."""

from __future__ import absolute_import
import unittest
from mock import patch, mock_open, MagicMock as Mock

from DIRAC import S_OK, S_ERROR
from Tests.Utilities.GeneralUtils import assertDiracFailsWith, assertDiracSucceedsWith, assertMockCalls

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.Workflow.Modules.RootExecutableAnalysis'


class TestRootExecutableAnalysis(unittest.TestCase):
  """Test the RootExecutableAnalysis module."""

  def setUp(self):
    from ILCDIRAC.Workflow.Modules.RootExecutableAnalysis import RootExecutableAnalysis
    self.rea = RootExecutableAnalysis()

  def test_applicationspecificinputs_noscript(self):
    assertDiracFailsWith(self.rea.applicationSpecificInputs(), 'script no defined', self)

  def test_applicationspecificinputs(self):
    self.rea.script = 'myTestscript'
    assertDiracSucceedsWith(self.rea.applicationSpecificInputs(), 'Parameters resolved', self)

  def test_runit_complete(self):
    self.rea.platform = 'myTestPlatform'
    self.rea.applicationLog = '/my/applog/test.log'
    self.rea.script = '/my/test/my_testscript123.sh'
    self.rea.workflowStatus['OK'] = True
    self.rea.stepStatus['OK'] = True
    self.rea.applicationVersion = 'v123'
    self.rea.STEP_NUMBER = 13
    self.rea.arguments = 'mytest args 123'
    self.rea.extraCLIarguments = 'extra test_args'
    with patch('%s.getEnvironmentScript' % MODULE_NAME, new=Mock(return_value=S_OK('/mytestenvscript'))), \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=[False, False, False, False, True])) as exists_mock, \
         patch('%s.os.remove' % MODULE_NAME, new=Mock()) as remove_mock, \
         patch('%s.open' % MODULE_NAME, mock_open(), create=True) as open_mock, \
         patch('%s.os.chmod' % MODULE_NAME, new=Mock()) as chmod_mock, \
         patch('%s.shellCall' % MODULE_NAME, new=Mock(return_value=S_OK((0, 'my_statusmsg')))) as cmd_mock:
      assertDiracSucceedsWith(self.rea.runIt(), 'ROOT v123 Successful', self)
      assertMockCalls(exists_mock, ['Root_v123_Run_13.sh', './lib', 'my_testscript123.sh',
                                      '/my/applog/test.log', '/my/applog/test.log'], self)
      self.assertFalse(remove_mock.called)
      open_mock.assert_called_once_with('Root_v123_Run_13.sh', 'w')
      open_mock = open_mock()
      open_mock.close.assert_called_once_with()
      assertMockCalls(open_mock.write, [
          '#!/bin/sh \n', '#####################################################################\n',
          '# Dynamically generated script to run a production or analysis job. #\n',
          '#####################################################################\n',
          'source /mytestenvscript\n', 'echo =============================\n', 'echo LD_LIBRARY_PATH is\n',
          'echo $LD_LIBRARY_PATH | tr ":" "\n"\n', 'echo =============================\n',
          'echo PATH is\n', 'echo $PATH | tr ":" "\n"\n', 'echo =============================\n',
          'env | sort >> localEnv.log\n', 'echo =============================\n',
          'my_testscript123.sh mytest args 123 extra test_args\n', 'declare -x appstatus=$?\n',
          'exit $appstatus\n'], self)
      chmod_mock.assert_called_once_with('Root_v123_Run_13.sh', 0o755)
      cmd_mock.assert_called_once_with(0, 'sh -c "./Root_v123_Run_13.sh"',
                                        callbackFunction=self.rea.redirectLogOutput,
                                        bufferLimit=20971520)

  def test_runit_noplatform(self):
    self.rea.applicationLog = 'asd'
    assertDiracFailsWith(self.rea.runIt(), 'no ilc platform selected', self)

  def test_runit_nolog(self):
    self.rea.platform = 'asd'
    assertDiracFailsWith(self.rea.runIt(), 'no log file provided', self)

  def test_runit_getenvscript_fails(self):
    self.rea.platform = 'myTestPlatform'
    self.rea.applicationLog = '/my/applog/test.log'
    self.rea.script = '/my/test/script.sh'
    with patch('%s.getEnvironmentScript' % MODULE_NAME, new=Mock(return_value=S_ERROR('test_scripterr'))):
      assertDiracFailsWith(self.rea.runIt(), 'test_scripterr', self)

  def test_runit_status_not_ok(self):
    self.rea.platform = 'myTestPlatform'
    self.rea.applicationLog = '/my/applog/test.log'
    self.rea.script = '/my/test/script.sh'
    self.rea.workflowStatus['OK'] = True
    self.rea.stepStatus['OK'] = False
    with patch('%s.getEnvironmentScript' % MODULE_NAME, new=Mock(return_value=S_OK('/mytestenvscript'))):
      assertDiracSucceedsWith(self.rea.runIt(),
                               'ROOT should not proceed as previous step did not end properly', self)

  def test_runit_noscript(self):
    self.rea.platform = 'myTestPlatform'
    self.rea.applicationLog = '/my/applog/test.log'
    self.rea.script = ''
    self.rea.workflowStatus['OK'] = True
    self.rea.stepStatus['OK'] = True
    with patch('%s.getEnvironmentScript' % MODULE_NAME, new=Mock(return_value=S_OK('/mytestenvscript'))):
      assertDiracFailsWith(self.rea.runIt(), 'executable file not defined', self)

  def test_runit_no_applog_created(self):
    self.rea.platform = 'myTestPlatform'
    self.rea.applicationLog = '/my/applog/test.log'
    self.rea.script = '/my/test/my_testscript123.sh'
    self.rea.workflowStatus['OK'] = True
    self.rea.stepStatus['OK'] = True
    self.rea.applicationVersion = 'v123'
    self.rea.STEP_NUMBER = 13
    self.rea.arguments = 'mytest args 123'
    self.rea.extraCLIarguments = 'extra test_args'
    with patch('%s.getEnvironmentScript' % MODULE_NAME, new=Mock(return_value=S_OK('/mytestenvscript'))), \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=[True, True, True, True, False])) as exists_mock, \
         patch('%s.os.remove' % MODULE_NAME, new=Mock()) as remove_mock, \
         patch('%s.open' % MODULE_NAME, mock_open(), create=True) as open_mock, \
         patch('%s.os.chmod' % MODULE_NAME, new=Mock()) as chmod_mock, \
         patch('%s.shellCall' % MODULE_NAME, new=Mock(return_value=S_OK(('something', True)))) as cmd_mock:
      assertDiracFailsWith(self.rea.runIt(), 'root did not produce the expected log', self)
      assertMockCalls(exists_mock, ['Root_v123_Run_13.sh', './lib', 'my_testscript123.sh',
                                      '/my/applog/test.log', '/my/applog/test.log'], self)
      assertMockCalls(remove_mock, ['Root_v123_Run_13.sh', '/my/applog/test.log'], self)
      open_mock.assert_called_once_with('Root_v123_Run_13.sh', 'w')
      open_mock = open_mock()
      open_mock.close.assert_called_once_with()
      assertMockCalls(open_mock.write, [
          '#!/bin/sh \n', '#####################################################################\n',
          '# Dynamically generated script to run a production or analysis job. #\n',
          '#####################################################################\n',
          'source /mytestenvscript\n', 'declare -x LD_LIBRARY_PATH=./lib:$LD_LIBRARY_PATH\n',
          'echo =============================\n', 'echo LD_LIBRARY_PATH is\n',
          'echo $LD_LIBRARY_PATH | tr ":" "\n"\n', 'echo =============================\n',
          'echo PATH is\n', 'echo $PATH | tr ":" "\n"\n', 'echo =============================\n',
          'env | sort >> localEnv.log\n', 'echo =============================\n', 'chmod u+x my_testscript123.sh\n',
          './my_testscript123.sh mytest args 123 extra test_args\n', 'declare -x appstatus=$?\n',
          'exit $appstatus\n'], self)
      chmod_mock.assert_called_once_with('Root_v123_Run_13.sh', 0o755)
      cmd_mock.assert_called_once_with(0, 'sh -c "./Root_v123_Run_13.sh"',
                                        callbackFunction=self.rea.redirectLogOutput,
                                        bufferLimit=20971520)



#!/usr/bin/env python
#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Created on Feb 8, 2012.

:author: Stephane Poss
"""
from __future__ import print_function
from __future__ import absolute_import
__RCSID__ = "$Id$"
# pylint: disable=C0103

from DIRAC.Core.Base import Script

from DIRAC import S_OK, S_ERROR


# As it's a full chain, we start at generation
# so we need to define the process and the energy
# The rest will be set later. We could also set the process
# and the energy directly in the whizard def, but for clarity
# it's better to do it before, that way we know the very
# essential

class Params(object):
  """command line parameters to create test productions."""

  def __init__(self):
    self.knownModels = ['CLIC_ILD_CDR', 'ILD_o1_v05', "CLIC_o2_v03"]

    self.energy = 250
    self.detectorModel = "CLIC_o2_v03"
    self.outputSE = "CERN-DIP-4"
    self.ildConfig = "DDSim"
    self.onlyDestination = []
    self.logLevel = "INFO"
    self.numberOfEvents = 10
    self.dryrun = False

  def setDetectorModel(self, model):
    if model not in self.knownModels:
      return S_ERROR("Unknown detectormodel, use one of these %s " % (",".join(self.knownModels)))
    self.detectorModel = model
    return S_OK()

  def setOutputSE(self, se):
    self.outputSE = se
    return S_OK()

  def setILDConfig(self, config):
    self.ildConfig = config
    return S_OK()

  def setOnlyDestination(self, dest):
    if isinstance(dest, list):
      self.onlyDestination = dest
    elif ',' in dest:
      self.onlyDestination = dest.split(',')
    elif ' ' in dest:
      self.onlyDestination = dest.split(' ')
    return S_OK()

  def setEnergy(self, energy):
    self.energy = energy
    return S_OK()

  def setDryRun(self, _dummy):
    self.dryrun = True
    return S_OK()

  def setNumberOfEvents(self, numberOfEvents):
    try:
      numberOfEvents = int(numberOfEvents)
    except ValueError as e:
      return S_ERROR("ERROR: Number of events needs to be an integer: %s" % str(e))
    self.numberOfEvents = numberOfEvents
    return S_OK()

  def registerSwitches(self):
    Script.registerSwitch("O:", "OutputSE=", "Output SE", self.setOutputSE)
    Script.registerSwitch("D:", "Destination=", "Only send jobs to these CEs", self.setOnlyDestination)
    Script.registerSwitch(
        "M:",
        "DetectorModel=",
        "DetectorModel: [%s]" %
        ", ".join(
            self.knownModels),
        self.setDetectorModel)
    Script.registerSwitch("C:", "Configuration=", "Configuration with Steeringfiles", self.setILDConfig)
    Script.registerSwitch("E:", "Energy=", "Energy", self.setEnergy)
    Script.registerSwitch("N:", "NumberOfEvents=", "Number of Events", self.setNumberOfEvents)
    Script.registerSwitch("", "dry-run", "Just do a dry run", self.setDryRun)
    Script.setUsageMessage("%s [opts] <extraName>" % Script.scriptName)


def getdicts(process):
  """Create the proper structures to build all the prodcutions for the samples with ee_, ea_ aa_."""
  plist = []
  if process.count("ee_"):
    plist.append({'process': process, 'pname1': 'e1', 'pname2': 'E1',
                 "epa_b1": 'F', "epa_b2": 'F', "isr_b1": 'T', "isr_b2": 'T'})
  elif process.count("ea_"):
    plist.append({'process': process, 'pname1': 'e1', 'pname2': 'E1',
                 "epa_b1": 'F', "epa_b2": 'T', "isr_b1": 'T', "isr_b2": 'F'})
    plist.append({'process': process, 'pname1': 'e1', 'pname2': 'A',
                 "epa_b1": 'F', "epa_b2": 'F', "isr_b1": 'T', "isr_b2": 'F'})
    plist.append({'process': process.replace("ea_", "ae_"), 'pname1': 'e1', 'pname2': 'E1',
                 "epa_b1": 'T', "epa_b2": 'F', "isr_b1": 'F', "isr_b2": 'T'})
    plist.append({'process': process.replace("ea_", "ae_"), 'pname1': 'A', 'pname2': 'E1',
                 "epa_b1": 'F', "epa_b2": 'F', "isr_b1": 'F', "isr_b2": 'T'})
  elif process.count("aa_"):
    plist.append({'process': process, 'pname1': 'e1', 'pname2': 'E1',
                 "epa_b1": 'T', "epa_b2": 'T', "isr_b1": 'F', "isr_b2": 'F'})
    plist.append({'process': process, 'pname1': 'e1', 'pname2': 'A',
                 "epa_b1": 'T', "epa_b2": 'F', "isr_b1": 'F', "isr_b2": 'F'})
    plist.append({'process': process, 'pname1': 'A', 'pname2': 'E1',
                 "epa_b1": 'F', "epa_b2": 'T', "isr_b1": 'F', "isr_b2": 'F'})
    plist.append({'process': process, 'pname1': 'A', 'pname2': 'A',
                 "epa_b1": 'F', "epa_b2": 'F', "isr_b1": 'F', "isr_b2": 'F'})
  else:
    plist.append({'process': process, 'pname1': 'e1', 'pname2': 'E1',
                 "epa_b1": 'F', "epa_b2": 'F', "isr_b1": 'T', "isr_b2": 'T'})
  return plist


PARAMS = Params()
PARAMS.registerSwitches()
Script.parseCommandLine()
extraargs = Script.getPositionalArgs()
if len(extraargs) == 0:
  print("ERROR: ExtraName not defined")
  Script.showHelp()
  raise RuntimeError("1")
additional_name = extraargs[0]


from ILCDIRAC.Interfaces.API.NewInterface.ProductionJob import ProductionJob
from ILCDIRAC.Interfaces.API.NewInterface.Applications import Whizard, Mokka, Marlin, OverlayInput, StdhepCut, StdhepCutJava, DDSim
from ILCDIRAC.Interfaces.API.NewInterface.Applications import SLIC, LCSIM, SLICPandora, SLCIOSplit, StdHepSplit
from ILCDIRAC.Interfaces.API.DiracILC import DiracILC


energy = float(PARAMS.energy)


## tripleH, Hrecoil, stau, gauginos, Hmass, tt, Htautau, Hmumu, Hee, Hbbccmumu, squarks, LCFITraining, Hgammagamma
# HZgamma Hinclusive ZZfusion, Any, ttH, bb_cc_gg
analysis = 'several'
process = 'hzqq'
#additional_name = '_neu1_356'
globname = ""
meta_energy = str(int(energy))

# For meta def
meta = {}
meta['ProdID'] = 1
meta['EvtType'] = process
meta['Energy'] = meta_energy

#ILDCONFIG = "CLICSteeringFilesV22"
ILDCONFIG = "v01-16-p10_250"
SOFTWAREVERSION = "ILCSoft-01-17-06"

ILDDetectorModels = ['ILD_o1_v05']
CLICDetectorModels = ['CLIC_ILD_CDR']

detectormodel = PARAMS.detectorModel

# Here get the prod list: initial particles combinasions
prodlist = getdicts(process)

beamrecoil = False

# Do generation
gen = True

# For cuts
cut = False
if cut:
  additional_name += "_cut"

javacut = False
cutfile = 'cuts_qq_nunu_1400.txt'
seleff = 0.52
n_keep = 500

# DoSplit
activesplitstdhep = False
if activesplitstdhep:
  additional_name += "_gensplit"
nbevtsperfilestdhep = 10

# Do Sim
ild_sim = False
sid_sim = False
dd_sim = True

# DoSplit
activesplit = False
if activesplit:
  additional_name += "_simsplit"

nbevtsperfile = 10

# Do Reco
ild_rec = False
sid_rec = False
# Do Reco with Overlay
ild_rec_ov = False
sid_rec_ov = False

n_events = PARAMS.numberOfEvents
#rodOutputSE = "PNNL3-SRM"
prodOutputSE = PARAMS.outputSE
onlyDestination = PARAMS.onlyDestination
logLevel = PARAMS.logLevel

model = 'sm'
#model = 'susyStagedApproach'
#model = 'susyStagedApproach'+additional_name
#additional_name = '_'
# NO TAUOLA ADD MSTJ(28)=2 in pythia parameters
whcutdict = {}
#whcutdict = {'e2e2_o':["cut M of 3 within 100 150"]}

additionnalreqs = ''
# additionnalreqs = 'MDCY(25,2)=214;MDCY(25,3)=1;' #for decay of H to bb
# additionnalreqs = 'MDCY(25,2)=213;MDCY(25,3)=1;' #for decay of H to cc
# additionnalreqs = 'MDCY(25,2)=219;MDCY(25,3)=1;' #for decay of H to mumu
# additionnalreqs = 'MWID(25)=2;MDCY(25,2)=223;MDCY(25,3)=1;BRAT(223)=1' #for decay of H to gammagamma
# additionnalreqs = 'MWID(25)=2;MDCY(25,2)=224;MDCY(25,3)=1;BRAT(224)=1;MDCY(23,2)=174;MDCY(23,3)=5' #for decay of H to Zgamma, Z to qq
# additionnalreqs = 'MWID(25)=2;MDCY(25,2)=224;MDCY(25,3)=1;BRAT(224)=1;MWID(23)=2;MDCY(23,2)=182;MDCY(23,3)=6; BRAT(182)=0.3333;BRAT(183)=0.;BRAT(184)=0.3333;BRAT(185)=0.;BRAT(186)=0.333$
# additionnalreqs = 'MDCY(25,3)=5;' #for decay of H to quarks
# additionnalreqs = 'MDCY(25,2)=220; MDCY(25,3)=1;'#for decay of H to tau tau
# additionnalreqs = 'MDCY(23,2)=174; MDCY(23,3)=3;'#for decay of Z to bb
mh = 126.
mb = 2.9
mc = 0
mmu = 0.10566
mtau = 1.77
#alphas = 0.000001
# Use gridfiles ?
gridfiles = False
# Set exchange lines to 4 (needed for complex final states)
exchange_lines = False


dirac = DiracILC()


# Whatever is below is not to be touched...
for proddict in prodlist:
  prod_name = additional_name
  process = proddict['process']

  if energy == 1400.:
    spectrum = 19
  elif energy == 3000.:
    spectrum = 11
  elif energy == 500.:
    spectrum = 13
  elif energy == 420.:
    spectrum = 13
  elif energy == 250.:
    spectrum = 20
  elif energy == 350.:
    spectrum = 20
  else:
    print("No spectrum defined, cannot proceed")
    raise RuntimeError("1")

  # Start by defining the whizard application

  wh = Whizard(processlist=dirac.getProcessList())
  wh.setModel(model)
  pdict = {}
  pdict['process_input'] = {}
  pdict['process_input']['process_id'] = proddict['process']
  pdict['process_input']['sqrts'] = energy
  if not model == 'sm':
    pdict['process_input']['input_file'] = "LesHouches.msugra_1"
    pdict['process_input']['input_slha_format'] = 'T'
  if beamrecoil and gen:
    prod_name += "_beamrecoil"
    pdict['process_input']['beam_recoil'] = 'T'

  pdict['integration_input'] = {}
  pdict['integration_input']['calls'] = '1  500000 10  500000  1  1500000'
  if gridfiles:
    pdict['integration_input']['read_grids'] = 'T'
  if exchange_lines:
    pdict['integration_input']['exchange_lines'] = 4
  pdict['simulation_input'] = {}
  pdict['simulation_input']['normalize_weight'] = 'F'
  pdict['simulation_input']['n_events'] = n_events
  pdict['simulation_input']['keep_initials'] = 'T'
  pdict['simulation_input']['events_per_file'] = 500000
  if model != 'sm':
    pdict['simulation_input']['pythia_parameters'] = "PMAS(25,1)=%s; PMAS(25,2)=0.3605E-02; MSTU(22)=20 ;PARJ(21)=0.40000;PARJ(41)=0.11000; PARJ(42)=0.52000; PARJ(81)=0.25000; PARJ(82)=1.90000; MSTJ(11)=3; PARJ(54)=-0.03100; PARJ(55)=-0.00200;PARJ(1)=0.08500; PARJ(3)=0.45000; PARJ(4)=0.02500; PARJ(2)=0.31000; PARJ(11)=0.60000; PARJ(12)=0.40000; PARJ(13)=0.72000;PARJ(14)=0.43000; PARJ(15)=0.08000; PARJ(16)=0.08000; PARJ(17)=0.17000; MSTP(3)=1;IMSS(1)=11; IMSS(21)=71; IMSS(22)=71;%s" % (mh, additionnalreqs)
  else:
    pdict['simulation_input']['pythia_parameters'] = "PMAS(25,1)=%s; PMAS(25,2)=0.3605E-02; MSTU(22)=20 ; MSTJ(28)=2 ;PARJ(21)=0.40000;PARJ(41)=0.11000; PARJ(42)=0.52000; PARJ(81)=0.25000; PARJ(82)=1.90000; MSTJ(11)=3; PARJ(54)=-0.03100; PARJ(55)=-0.00200;PARJ(1)=0.08500; PARJ(3)=0.45000; PARJ(4)=0.02500; PARJ(2)=0.31000; PARJ(11)=0.60000; PARJ(12)=0.40000; PARJ(13)=0.72000;PARJ(14)=0.43000; PARJ(15)=0.08000; PARJ(16)=0.08000; PARJ(17)=0.17000; MSTP(3)=1;%s" % (mh, additionnalreqs)
    pdict['parameter_input'] = {}
  #  pdict['parameter_input']['mmu']=mmu
  #  pdict['parameter_input']['mtau']=mtau
  #  pdict['parameter_input']['mb']=mb
  #  pdict['parameter_input']['mc']=mc
    pdict['parameter_input']['mH'] = mh
  #  pdict['parameter_input']['alphas']=alphas
  pdict['beam_input_1'] = {}
  pdict['beam_input_1']['particle_name'] = proddict['pname1']
  pdict['beam_input_1']['polarization'] = "0.0 0.0"
  pdict['beam_input_1']['USER_spectrum_on'] = 'T'
  pdict['beam_input_1']['USER_spectrum_mode'] = spectrum
  pdict['beam_input_1']['ISR_on'] = proddict['isr_b1']
  pdict['beam_input_1']['EPA_on'] = proddict['epa_b1']

  pdict['beam_input_2'] = {}
  pdict['beam_input_2']['particle_name'] = proddict['pname2']
  pdict['beam_input_2']['polarization'] = "0.0 0.0"
  pdict['beam_input_2']['USER_spectrum_on'] = 'T'
  pdict['beam_input_2']['ISR_on'] = proddict['isr_b2']
  pdict['beam_input_2']['USER_spectrum_mode'] = -spectrum
  pdict['beam_input_2']['EPA_on'] = proddict['epa_b2']

  prod_name += "_" + proddict['pname1'] + proddict['epa_b1'] + "_" + proddict['pname2'] + proddict['epa_b2']

  wh.setFullParameterDict(pdict)
  # wh.setGlobalEvtType("aa_e3e3nn")
  #process = "aa_e3e3nn"
  if globname:
    wh.setGlobalEvtType(globname)
    process = globname

  if whcutdict:
    wh.setGeneratorLevelCuts(whcutdict)

  if cut:
    wh.willCut()

  if gridfiles:
    wh.usingGridFiles()

  if javacut:
    stdhepc = StdhepCutJava()
    stdhepc.setVersion('1.0')
  else:
    stdhepc = StdhepCut()
    stdhepc.setVersion("V7")
  if cut and not cutfile:
    print("No cut file defined, cannot proceed")
    raise RuntimeError("1")
  stdhepc.setSteeringFile(cutfile)
  stdhepc.setMaxNbEvts(n_keep)
  stdhepc.setSelectionEfficiency(seleff)

  # Split
  stdhepsplit = StdHepSplit()
  stdhepsplit.setVersion("V2")
  stdhepsplit.setNumberOfEventsPerFile(nbevtsperfilestdhep)

  # Simulation ILD
  mo = Mokka()
  mo.setVersion(SOFTWAREVERSION)
  # mo.setNbEvts(10)
  if energy in [500., 420., 375., 350., 250.]:
    mo.setSteeringFile("clic_ild_cdr500.steer")
  elif energy in [3000., 1400.]:
    mo.setSteeringFile("clic_ild_cdr.steer")
  else:
    print('Detector Model for Mokka undefined for this energy')
  if detectormodel == 'ild_00':
    mo.setSteeringFile("ild_00.steer")
  if detectormodel == 'ILD_o1_v05':
    mo.setSteeringFile("bbudsc_3evt.steer")

  # Simulation DDSim
  dd = DDSim()
  dd.setVersion("testVersion")
  dd.setNbEvts(10)
  # if energy in [500., 420., 375., 350., 250.,3000., 1400.]:
  #   mo.setSteeringFile("dssimsteering.py")
  # else:
  #   print 'SteeringFile Model for DDSim undefined for this energy'
  dd.setDetectorModel(detectormodel)

  # Simulation SID
  slic = SLIC()
  slic.setVersion('v2r9p8')
  slic.setSteeringFile('defaultClicCrossingAngle.mac')
  slic.setDetectorModel('clic_sid_cdr')

  # Split
  split = SLCIOSplit()
  split.setNumberOfEventsPerFile(nbevtsperfile)

  overlay = OverlayInput()
  if detectormodel in CLICDetectorModels:
    overlay.setMachine("clic_cdr")
    overlay.setEnergy(energy)
    overlay.setBkgEvtType("gghad")

    if energy == 500.:
      overlay.setBXOverlay(300)
      overlay.setGGToHadInt(0.3)  # When running at 500geV
      overlay.setDetectorModel("CLIC_ILD_CDR500")
    elif energy == 420.:
      overlay.setBXOverlay(300)
      overlay.setGGToHadInt(0.17)  # When running at 420geV
      overlay.setDetectorModel("CLIC_ILD_CDR500")
    elif energy == 350.:
      overlay.setBXOverlay(300)
      overlay.setGGToHadInt(0.0464)  # When running at 350geV
      overlay.setDetectorModel("CLIC_ILD_CDR500")
    elif energy == 3000.:
      overlay.setBXOverlay(60)
      overlay.setGGToHadInt(3.2)  # When running at 3TeV
      overlay.setDetectorModel("CLIC_ILD_CDR")
    elif energy == 1400.:
      overlay.setBXOverlay(60)
      overlay.setGGToHadInt(1.3)  # When running at 1.4TeV
      overlay.setDetectorModel("CLIC_ILD_CDR")
    else:
      print("Overlay CLIC_ILD: No overlay parameters defined for this energy")

  if detectormodel in ILDDetectorModels:
    overlay.setMachine("ilc_dbd")
    overlay.setBackgroundType("aa_lowpt")
    overlay.setBXOverlay(1)
    overlay.setDetectorModel(detectormodel)
    if energy == 250:
      overlay.setGGToHadInt(0.3)

  # Reconstruction ILD with overlay
  mao = Marlin()
  mao.setDebug()
  mao.setVersion(SOFTWAREVERSION)
  if ild_rec_ov:
    if energy == 500.:
      mao.setSteeringFile("clic_ild_cdr500_steering_overlay.xml")
      mao.setGearFile('clic_ild_cdr500.gear')
    elif energy == 420.:
      mao.setSteeringFile("clic_ild_cdr500_steering_overlay_420.0.xml")
      mao.setGearFile('clic_ild_cdr500.gear')
    elif energy == 350.:
      mao.setSteeringFile("clic_ild_cdr500_steering_overlay_350.0.xml")
      mao.setGearFile('clic_ild_cdr500.gear')
    elif energy == 3000.0:
      mao.setSteeringFile("clic_ild_cdr_steering_overlay_3000.0.xml")
      mao.setGearFile('clic_ild_cdr.gear')
    elif energy == 1400.0:
      mao.setSteeringFile("clic_ild_cdr_steering_overlay_1400.0.xml")
      mao.setGearFile('clic_ild_cdr.gear')
    else:
      print("Marlin: No reconstruction suitable for this energy")

  if detectormodel in ILDDetectorModels:
    mao.setSteeringFile("bbudsc_3evt_stdreco.xml")
    mao.setGearFile("GearOutput.xml")

  # Reconstruction w/o overlay
  ma = Marlin()
  ma.setDebug()
  ma.setVersion(SOFTWAREVERSION)
  if ild_rec:
    if energy in [500., 420., 350., 250.]:
      ma.setSteeringFile("clic_ild_cdr500_steering.xml")
      ma.setGearFile('clic_ild_cdr500.gear')
    elif energy in [3000., 1400.]:
      ma.setSteeringFile("clic_ild_cdr_steering.xml")
      ma.setGearFile('clic_ild_cdr.gear')
    else:
      print("Marlin: No reconstruction suitable for this energy")

  # SID Reco w/o overlay
  lcsim_prepandora = LCSIM()
  lcsim_prepandora.setVersion('CLIC_CDR')
  lcsim_prepandora.setSteeringFile("clic_cdr_prePandora.lcsim")
  lcsim_prepandora.setTrackingStrategy("defaultStrategies_clic_sid_cdr.xml")
  # lcsim_prepandora.setDetectorModel('clic_sid_cdr')
  lcsim_prepandora.setOutputFile("prePandora.slcio")
  lcsim_prepandora.willRunSLICPandora()

  slicpandora = SLICPandora()
  slicpandora.setVersion('CLIC_CDR')
  slicpandora.setDetectorModel('clic_sid_cdr')
  slicpandora.setPandoraSettings("PandoraSettingsSlic.xml")
  slicpandora.getInputFromApp(lcsim_prepandora)
  slicpandora.setOutputFile('pandora.slcio')

  lcsim_postpandora = LCSIM()
  lcsim_postpandora.setVersion('CLIC_CDR')
  lcsim_postpandora.getInputFromApp(slicpandora)
  lcsim_postpandora.setSteeringFile("clic_cdr_postPandoraOverlay.lcsim")
  lcsim_postpandora.setTrackingStrategy("defaultStrategies_clic_sid_cdr.xml")
  # lcsim_postpandora.setDetectorModel('clic_sid_cdr')

  # SID Reco w/o overlay
  overlay_sid = OverlayInput()
  overlay_sid.setMachine("clic_cdr")
  overlay_sid.setEnergy(energy)
  overlay_sid.setBkgEvtType("gghad")
  if energy == 3000.:
    overlay_sid.setBXOverlay(60)
    overlay_sid.setGGToHadInt(3.2)  # When running at 3TeV
  elif energy == 350.:
    overlay.setBXOverlay(300)
    overlay.setGGToHadInt(0.0464)  # When running at 350geV
  elif energy == 1400.:
    overlay_sid.setBXOverlay(60)
    overlay_sid.setGGToHadInt(1.3)  # When running at 1.4TeV
  else:
    print("Overlay SID: No overlay parameters defined for this energy")
  overlay_sid.setDetectorModel("CLIC_SID_CDR")

  lcsim_prepandora_ov = LCSIM()
  lcsim_prepandora_ov.setVersion('CLIC_CDR')
  if energy == 3000.0:
    lcsim_prepandora_ov.setSteeringFile("clic_cdr_prePandoraOverlay_3000.0.lcsim")
  elif energy == 1400.0:
    lcsim_prepandora_ov.setSteeringFile("clic_cdr_prePandoraOverlay_1400.0.lcsim")
  else:
    print("LCSIM: No steering files defined for this energy")

  lcsim_prepandora_ov.setTrackingStrategy("defaultStrategies_clic_sid_cdr.xml")
  # lcsim_prepandora_ov.setDetectorModel('clic_sid_cdr')
  lcsim_prepandora_ov.setOutputFile("prePandora.slcio")
  lcsim_prepandora_ov.willRunSLICPandora()

  slicpandora_ov = SLICPandora()
  slicpandora_ov.getInputFromApp(lcsim_prepandora_ov)
  slicpandora_ov.setVersion('CLIC_CDR')
  slicpandora_ov.setDetectorModel('clic_sid_cdr')
  slicpandora_ov.setPandoraSettings("PandoraSettingsSlic.xml")
  slicpandora_ov.setOutputFile('pandora.slcio')

  lcsim_postpandora_ov = LCSIM()
  lcsim_postpandora_ov.getInputFromApp(slicpandora_ov)
  lcsim_postpandora_ov.setVersion('CLIC_CDR')
  lcsim_postpandora_ov.setSteeringFile("clic_cdr_postPandoraOverlay.lcsim")
  lcsim_postpandora_ov.setTrackingStrategy("defaultStrategies_clic_sid_cdr.xml")
  # lcsim_postpandora_ov.setDetectorModel('clic_sid_cdr')

  if gen:
    ##########################################
    # Define the generation production.
    pwh = ProductionJob()
    pwh.setDryRun(PARAMS.dryrun)
    pwh.setLogLevel(logLevel)
    pwh.setOutputSE(prodOutputSE)
    if onlyDestination:
      pwh.setDestination(onlyDestination)
    pwh.setProdType("MCGeneration")
    wname = process + "_" + str(energy)
    if additionnalreqs:
      wname += "_forced_decay"
    if cut:
      wname += "_cut"
    wname += prod_name
    pwh.setWorkflowName(wname)
    pwh.setProdGroup(analysis + "_" + str(energy))
    res = pwh.append(wh)
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")

    if cut:
      res = pwh.append(stdhepc)
      if not res['OK']:
        print(res['Message'])
        raise RuntimeError("1")

    pwh.addFinalization(True, True, True, True)
    descrp = "CLIC %s BeamSpread, ISR ON, whizard" % energy
    if additionnalreqs:
      descrp += ", %s" % additionnalreqs
    if cut:
      descrp += ", cut at stdhep level"
    if prod_name:
      descrp += ", %s" % prod_name

    pwh.setDescription(descrp)

    res = pwh.createProduction()
    if not res['OK']:
      print(res['Message'])

    pwh.addMetadataToFinalFiles({"BeamParticle1": proddict['pname1'], "BeamParticle2": proddict['pname2'],
                                 "EPA_B1": proddict['epa_b1'], "EPA_B2": proddict['epa_b2']})

    res = pwh.finalizeProd()
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    pwh.setNbOfTasks(1)
    # The production is created, one can now take care of the second step:
    # For that we will use the metadata of the previous production as input
    meta = pwh.getMetadata()

  if activesplitstdhep and meta:
    pstdhepsplit = ProductionJob()
    pstdhepsplit.setDryRun(PARAMS.dryrun)
    pstdhepsplit.setLogLevel(logLevel)
    pstdhepsplit.setProdType('Split')
    if onlyDestination:
      pstdhepsplit.setDestination(onlyDestination)
    res = pstdhepsplit.setInputDataQuery(meta)
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    pstdhepsplit.setOutputSE(prodOutputSE)
    wname = process + "_" + str(energy) + "_split"
    wname += prod_name
    pstdhepsplit.setWorkflowName(wname)
    pstdhepsplit.setProdGroup(analysis + "_" + str(energy))

    # Add the application
    res = pstdhepsplit.append(stdhepsplit)
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    pstdhepsplit.addFinalization(True, True, True, True)
    descrp = "Splitting stdhep files"

    if prod_name:
      descrp += ", %s" % prod_name

    pstdhepsplit.setDescription(descrp)

    res = pstdhepsplit.createProduction()
    if not res['OK']:
      print(res['Message'])

    pstdhepsplit.addMetadataToFinalFiles({"BeamParticle1": proddict['pname1'], "BeamParticle2": proddict['pname2'],
                                          "EPA_B1": proddict['epa_b1'], "EPA_B2": proddict['epa_b2']})

    res = pstdhepsplit.finalizeProd()
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    # As before: get the metadata for this production to input into the next
    meta = pstdhepsplit.getMetadata()

  if ild_sim and meta:
    ####################
    # Define the second production (simulation). Notice the setInputDataQuery call
    pmo = ProductionJob()
    pmo.setDryRun(PARAMS.dryrun)
    pmo.setLogLevel(logLevel)
    if onlyDestination:
      pmo.setDestination(onlyDestination)
    pmo.setProdType('MCSimulation')
    pmo.setConfig(ILDCONFIG)
    res = pmo.setInputDataQuery(meta)
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    pmo.setOutputSE(prodOutputSE)
    wname = process + "_" + str(energy) + "_ild_sim"
    wname += prod_name
    pmo.setWorkflowName(wname)
    pmo.setProdGroup(analysis + "_" + str(energy))
    # Add the application
    res = pmo.append(mo)
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    pmo.addFinalization(True, True, True, True)
    if energy > 550.:
      descrp = "CLIC_ILD_CDR model"
    else:
      descrp = "CLIC_ILD_CDR_500 model"
    if prod_name:
      descrp += ", %s" % prod_name
    pmo.setDescription(descrp)
    res = pmo.createProduction()
    if not res['OK']:
      print(res['Message'])
    pmo.addMetadataToFinalFiles({"BeamParticle1": proddict['pname1'], "BeamParticle2": proddict['pname2'],
                                 "EPA_B1": proddict['epa_b1'], "EPA_B2": proddict['epa_b2']})
    res = pmo.finalizeProd()
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    # As before: get the metadata for this production to input into the next
    meta = pmo.getMetadata()

  if dd_sim and meta:
    ####################
    # Define the second production (simulation). Notice the setInputDataQuery call
    pdd = ProductionJob()
    pdd.setDryRun(PARAMS.dryrun)
    pdd.setLogLevel(logLevel)
    if onlyDestination:
      pdd.setDestination(onlyDestination)
    pdd.setProdType('MCSimulation')
    pdd.setConfig(ILDCONFIG)
    res = pdd.setInputDataQuery(meta)
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    pdd.setOutputSE(prodOutputSE)
    wname = process + "_" + str(energy) + "_ild_sim"
    wname += prod_name
    pdd.setWorkflowName(wname)
    pdd.setProdGroup(analysis + "_" + str(energy))
    # Add the application
    res = pdd.append(dd)
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    pdd.addFinalization(True, True, True, True)
    descrp = "DDSim model"
    if prod_name:
      descrp += ", %s" % prod_name
    pdd.setDescription(descrp)
    res = pdd.createProduction()
    if not res['OK']:
      print(res['Message'])
    pdd.addMetadataToFinalFiles({"BeamParticle1": proddict['pname1'], "BeamParticle2": proddict['pname2'],
                                 "EPA_B1": proddict['epa_b1'], "EPA_B2": proddict['epa_b2']})
    res = pdd.finalizeProd()
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    # As before: get the metadata for this production to input into the next
    meta = pdd.getMetadata()

  if sid_sim and meta:
    ####################
    # Define the second production (simulation). Notice the setInputDataQuery call
    psl = ProductionJob()
    psl.setDryRun(PARAMS.dryrun)
    psl.setLogLevel(logLevel)
    psl.setProdType('MCSimulation')
    if onlyDestination:
      psl.setDestination(onlyDestination)
    res = psl.setInputDataQuery(meta)
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    psl.setOutputSE(prodOutputSE)
    wname = process + "_" + str(energy) + "_sid_sim"
    wname += prod_name
    psl.setWorkflowName(wname)
    psl.setProdGroup(analysis + "_" + str(energy))
    # Add the application
    res = psl.append(slic)
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    psl.addFinalization(True, True, True, True)
    descrp = "CLIC_SID_CDR model"
    if prod_name:
      descrp += ", %s" % prod_name
    psl.setDescription(descrp)

    res = psl.createProduction()
    if not res['OK']:
      print(res['Message'])

    psl.addMetadataToFinalFiles({"BeamParticle1": proddict['pname1'], "BeamParticle2": proddict['pname2'],
                                 "EPA_B1": proddict['epa_b1'], "EPA_B2": proddict['epa_b2']})

    res = psl.finalizeProd()
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    # As before: get the metadata for this production to input into the next
    meta = psl.getMetadata()

  if activesplit and meta:
    #######################
    # Split the input files.
    psplit = ProductionJob()
    psplit.setDryRun(PARAMS.dryrun)
    psplit.setCPUTime(30000)
    if onlyDestination:
      psplit.setDestination(onlyDestination)
    psplit.setLogLevel(logLevel)
    psplit.setProdType('Split')
    psplit.setDestination("LCG.CERN.ch")
    res = psplit.setInputDataQuery(meta)
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    psplit.setOutputSE(prodOutputSE)
    wname = process + "_" + str(energy) + "_split"
    wname += prod_name
    psplit.setWorkflowName(wname)
    psplit.setProdGroup(analysis + "_" + str(energy))

    # Add the application
    res = psplit.append(split)
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    psplit.addFinalization(True, True, True, True)
    descrp = "Splitting slcio files"
    if prod_name:
      descrp += ", %s" % prod_name
    psplit.setDescription(descrp)

    res = psplit.createProduction()
    if not res['OK']:
      print(res['Message'])
    psplit.addMetadataToFinalFiles({"BeamParticle1": proddict['pname1'], "BeamParticle2": proddict['pname2'],
                                    "EPA_B1": proddict['epa_b1'], "EPA_B2": proddict['epa_b2']})

    res = psplit.finalizeProd()
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    # As before: get the metadata for this production to input into the next
    meta = psplit.getMetadata()

  if ild_rec and meta:
    #######################
    # Define the reconstruction prod
    pma = ProductionJob()
    pma.setDryRun(PARAMS.dryrun)
    pma.setLogLevel(logLevel)
    pma.setProdType('MCReconstruction')
    pma.setConfig(ILDCONFIG)
    if onlyDestination:
      pma.setDestination(onlyDestination)
    res = pma.setInputDataQuery(meta)
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    pma.setOutputSE(prodOutputSE)
    wname = process + "_" + str(energy) + "_ild_rec"
    wname += prod_name
    pma.setWorkflowName(wname)
    pma.setProdGroup(analysis + "_" + str(energy))

    # Add the application
    res = pma.append(ma)
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    pma.addFinalization(True, True, True, True)
    if energy > 550.:
      descrp = "CLIC_ILD_CDR, No overlay"
    else:
      descrp = "CLIC_ILD_CDR 500 gev, No overlay"
    if prod_name:
      descrp += ", %s" % prod_name
    pma.setDescription(descrp)

    res = pma.createProduction()
    if not res['OK']:
      print(res['Message'])
    pma.addMetadataToFinalFiles({"BeamParticle1": proddict['pname1'], "BeamParticle2": proddict['pname2'],
                                 "EPA_B1": proddict['epa_b1'], "EPA_B2": proddict['epa_b2']})

    res = pma.finalizeProd()
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")

  if sid_rec and meta:
    #######################
    # Define the reconstruction prod
    psidrec = ProductionJob()
    psidrec.setDryRun(PARAMS.dryrun)
    psidrec.setLogLevel(logLevel)
    psidrec.setProdType('MCReconstruction')
    psidrec.setBannedSites(['LCG.Bristol.uk', 'LCG.RAL-LCG2.uk'])
    if onlyDestination:
      psidrec.setDestination(onlyDestination)
    res = psidrec.setInputDataQuery(meta)
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    psidrec.setOutputSE(prodOutputSE)
    wname = process + "_" + str(energy) + "_sid_rec"
    wname += prod_name
    psidrec.setWorkflowName(wname)
    psidrec.setProdGroup(analysis + "_" + str(energy))
    res = psidrec.append(lcsim_prepandora)
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    res = psidrec.append(slicpandora)
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    res = psidrec.append(lcsim_postpandora)
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    psidrec.addFinalization(True, True, True, True)
    descrp = "CLIC_SID_CDR, No overlay"
    if prod_name:
      descrp += ", %s" % prod_name
    psidrec.setDescription(descrp)

    res = psidrec.createProduction()
    if not res['OK']:
      print(res['Message'])

    psidrec.addMetadataToFinalFiles({"BeamParticle1": proddict['pname1'], "BeamParticle2": proddict['pname2'],
                                     "EPA_B1": proddict['epa_b1'], "EPA_B2": proddict['epa_b2']})

    res = psidrec.finalizeProd()
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")

  if ild_rec_ov and meta:
    #######################
    # Define the reconstruction prod
    pmao = ProductionJob()
    pmao.setDryRun(PARAMS.dryrun)
    pmao.setLogLevel(logLevel)
    pmao.setProdType('MCReconstruction_Overlay')
    if onlyDestination:
      pmao.setDestination(onlyDestination)
    res = pmao.setInputDataQuery(meta)
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    pmao.setOutputSE(prodOutputSE)
    wname = process + "_" + str(energy) + "_ild_rec_overlay"
    wname += prod_name
    pmao.setWorkflowName(wname)
    pmao.setProdGroup(analysis + "_" + str(energy))

    # Add the application
    res = pmao.append(overlay)
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    # Add the application
    res = pmao.append(mao)
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    pmao.addFinalization(True, True, True, True)
    if energy > 550.:
      descrp = "CLIC_ILD_CDR, Overlay"
    else:
      descrp = "CLIC_ILD_CDR_500, Overlay"
    if prod_name:
      descrp += ", %s" % prod_name
    pmao.setDescription(descrp)
    res = pmao.createProduction()
    if not res['OK']:
      print(res['Message'])

    pmao.addMetadataToFinalFiles({"BeamParticle1": proddict['pname1'], "BeamParticle2": proddict['pname2'],
                                  "EPA_B1": proddict['epa_b1'], "EPA_B2": proddict['epa_b2']})

    res = pmao.finalizeProd()
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")

  if sid_rec_ov and meta:
    #######################
    # Define the reconstruction prod
    psidreco = ProductionJob()
    psidreco.setDryRun(PARAMS.dryrun)
    psidreco.setLogLevel(logLevel)
    psidreco.setProdType('MCReconstruction_Overlay')
    psidreco.setBannedSites(['LCG.Bristol.uk', 'LCG.RAL-LCG2.uk'])
    if onlyDestination:
      psidreco.setDestination(onlyDestination)
    res = psidreco.setInputDataQuery(meta)
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    psidreco.setOutputSE(prodOutputSE)
    wname = process + "_" + str(energy) + "_sid_rec_overlay"
    wname += prod_name
    psidreco.setWorkflowName(wname)
    psidreco.setProdGroup(analysis + "_" + str(energy))
    res = psidreco.append(overlay_sid)
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    res = psidreco.append(lcsim_prepandora_ov)
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    res = psidreco.append(slicpandora_ov)
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    res = psidreco.append(lcsim_postpandora_ov)
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")
    psidreco.addFinalization(True, True, True, True)
    descrp = "CLIC_SID_CDR, overlay"
    if prod_name:
      descrp += ", %s" % prod_name
    psidreco.setDescription(descrp)

    res = psidreco.createProduction()
    if not res['OK']:
      print(res['Message'])
    psidreco.addMetadataToFinalFiles({"BeamParticle1": proddict['pname1'], "BeamParticle2": proddict['pname2'],
                                      "EPA_Beam1": proddict['epa_b1'], "EPA_Beam2": proddict['epa_b2']})

    res = psidreco.finalizeProd()
    if not res['OK']:
      print(res['Message'])
      raise RuntimeError("1")

  # In principle nothing else is needed.



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Module to upload specified job output files according to the parameters defined in the production workflow.

:author: S. Poss
:since: Sep 01, 2010
"""

from __future__ import absolute_import
import os
import random
import time
from pprint import pformat

from DIRAC.DataManagementSystem.Client.FailoverTransfer import FailoverTransfer
from DIRAC import S_OK, S_ERROR, gLogger, gConfig
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
import DIRAC

from ILCDIRAC.Core.Utilities.ResolveSE import getDestinationSEList
from ILCDIRAC.Core.Utilities.resolvePathsAndNames import getProdFilename
from ILCDIRAC.Workflow.Modules.ModuleBase import ModuleBase
from ILCDIRAC.Core.Utilities.ProductionData import getExperimentFromPath
import six

__RCSID__ = '$Id$'
LOG = gLogger.getSubLogger(__name__)


class UploadOutputData(ModuleBase):
  """As name suggest: upload output data.

  For Production only: See :mod:`~ILCDIRAC.Workflow.Modules.UserJobFinalization` for User job upload.
  """
  #############################################################################

  def __init__(self):
    """Module initialization."""
    super(UploadOutputData, self).__init__()
    self.version = __RCSID__
    self.commandTimeOut = 10 * 60
    self.enable = True
    self.failoverSEs = gConfig.getValue('/Resources/StorageElementGroups/Tier1-Failover', [])
    self.ops = Operations()

    # List all parameters here
    self.outputDataFileMask = ''
    self.outputMode = 'Any'  # or 'Local' for reco case
    self.outputList = []
    self.productionID = 0
    self.prodOutputLFNs = []
    self.experiment = "CLIC"

  #############################################################################
  def applicationSpecificInputs(self):
    """By convention the module parameters are resolved here."""
    LOG.debug("Workflow commons: %s" % pformat(self.workflow_commons))

    self.enable = self.step_commons.get('Enable', self.enable)
    if not isinstance(self.enable, bool):
      LOG.warn('Enable flag set to non-boolean value %s, setting to False' % self.enable)
      self.enable = False

    self.productionID = self.workflow_commons.get("PRODUCTION_ID", self.productionID)

    self.jobID = os.environ.get('JOBID', self.jobID)
    if self.jobID:
      LOG.verbose('Found WMS JobID = %s' % self.jobID)
    else:
      LOG.info('No WMS JobID found, disabling module via control flag')
      self.enable = False

    # This is the thing that is used to establish the list of outpufiles to treat:
    # Make sure that all that is in the : "outputList" and also in the ProductionData
    # is treated properly. Needed as whatever is in outputList does not contain any reference to the
    # prodID and task ID. Also if for some reason a step failed, then the corresponding data will not be there
    self.outputList = self.workflow_commons.get('outputList', self.outputList)
    if self.outputList:
      if 'ProductionOutputData' in self.workflow_commons:
        productionData = self.workflow_commons['ProductionOutputData'].split(";")
        LOG.verbose("prod data : %s" % productionData)
        treatedOutputlist = {}
        for expectedOutputfile in self.outputList:
          LOG.debug("Treating file: %s" % expectedOutputfile['outputFile'])
          self.getTreatedOutputlistNew(productionData, treatedOutputlist, expectedOutputfile)
        self.outputList = list(treatedOutputlist.values())
      else:
        olist = []
        for expectedOutputfile in self.outputList:
          appdict = expectedOutputfile
          appdict['outputFile'] = getProdFilename(expectedOutputfile['outputFile'],
                                                  int(self.workflow_commons["PRODUCTION_ID"]),
                                                  int(self.workflow_commons["JOB_ID"]))
          olist.append(appdict)
        self.outputList = olist

      LOG.verbose("OutputList : %s" % self.outputList)

    self.outputMode = self.workflow_commons.get('outputMode', self.outputMode)

    self.outputDataFileMask = self.workflow_commons.get('outputDataFileMask', self.outputDataFileMask)
    if not isinstance(self.outputDataFileMask, list):
      self.outputDataFileMask = [i.lower().strip() for i in self.outputDataFileMask.split(';')]

    #result = constructProductionLFNs(self.workflow_commons)
    # if not result['OK']:
    #  LOG.error('Could not create production LFNs',result['Message'])
    #  return result
    # self.prodOutputLFNs=result['Value']['ProductionOutputData']

    tempOutputLFNs = self.workflow_commons.get('ProductionOutputData', self.prodOutputLFNs)
    if isinstance(tempOutputLFNs, six.string_types):
      self.prodOutputLFNs = tempOutputLFNs.split(";")
    else:
      self.prodOutputLFNs = tempOutputLFNs
    self.setJobParameter('ProductionOutputData', ';'.join(self.prodOutputLFNs), sendFlag=True)

    return S_OK('Parameters resolved')

  #############################################################################
  def execute(self):
    """Main execution function."""
    LOG.info('Initializing %s' % self.version)
    result = self.resolveInputVariables()
    if not result['OK']:
      LOG.error("Failed to resolve input parameters:", result['Message'])
      return result

    if not self.workflowStatus['OK'] or not self.stepStatus['OK']:
      LOG.verbose('Workflow status = %s, step status = %s' % (self.workflowStatus['OK'], self.stepStatus['OK']))
      return S_OK('No output data upload attempted')

    self.experiment = getExperimentFromPath(LOG, self.prodOutputLFNs[0], self.experiment)

    # Determine the final list of possible output files for the
    # workflow and all the parameters needed to upload them.
    result = self.getCandidateFiles(self.outputList, self.prodOutputLFNs, self.outputDataFileMask)
    if not result['OK']:
      LOG.error(result['Message'])
      self.setApplicationStatus(result['Message'])
      return result

    fileDict = result['Value']
    result = self.getFileMetadata(fileDict)
    if not result['OK']:
      LOG.error(result['Message'])
      self.setApplicationStatus(result['Message'])
      return result

    if not result['Value']:
      LOG.info('No output data files were determined to be uploaded for this workflow')
      return S_OK()

    fileMetadata = result['Value']

    # Get final, resolved SE list for files
    final = {}
    for fileName, metadata in fileMetadata.items():
      result = getDestinationSEList(metadata['workflowSE'], DIRAC.siteName(), self.outputMode)
      if not result['OK']:
        LOG.error('Could not resolve output data SE', result['Message'])
        self.setApplicationStatus('Failed To Resolve OutputSE')
        return result

      resolvedSE = result['Value']
      final[fileName] = metadata
      final[fileName]['resolvedSE'] = resolvedSE

    LOG.info('The following files will be uploaded: %s' % (', '.join(list(final.keys()))))
    for fileName, metadata in final.items():
      LOG.info('--------%s--------' % fileName)
      for metaName, metaValue in metadata.items():
        LOG.info('%s = %s' % (metaName, metaValue))

    # At this point can exit and see exactly what the module would have uploaded
    if not self.enable:
      LOG.info('Module is disabled by control flag, would have attempted to upload the \
      following files %s' % ', '.join(list(final.keys())))
      return S_OK('Module is disabled by control flag')

    # Disable the watchdog check in case the file uploading takes a long time
    LOG.info('Creating DISABLE_WATCHDOG_CPU_WALLCLOCK_CHECK in order to disable the Watchdog prior to upload')
    fopen = open('DISABLE_WATCHDOG_CPU_WALLCLOCK_CHECK', 'w')
    fopen.write('%s' % time.asctime())
    fopen.close()

    # Instantiate the failover transfer client with the global request object
    failoverTransfer = FailoverTransfer(self._getRequestContainer())

    catalogs = self.ops.getValue('Production/%s/Catalogs' % self.experiment,
                                 ['FileCatalog', 'LcgFileCatalog'])

    self.failoverSEs = self.ops.getValue("Production/%s/FailOverSE" % self.experiment, self.failoverSEs)

    # One by one upload the files with failover if necessary
    cleanUp = False
    for fileName, metadata in final.items():
      LOG.info("Attempting to store file %s to the following SE(s):\n%s" % (fileName,
                                                                                 ', '.join(metadata['resolvedSE'])))
      resultPrimary = failoverTransfer.transferAndRegisterFile(fileName,
                                                               metadata['localpath'],
                                                               metadata['lfn'],
                                                               metadata['resolvedSE'],
                                                               fileMetaDict=metadata['filedict'],
                                                               fileCatalog=catalogs)

      if resultPrimary['OK']:
        LOG.info("Successfully uploaded %s to %s" % (fileName, ', '.join(metadata['resolvedSE'])))
        continue

      # do the failover transfer
      LOG.error('Could not transfer and register %s with metadata:\n %s' %
                     (fileName, pformat(metadata['filedict'])))
      failovers = self.failoverSEs
      targetSE = metadata['resolvedSE'][0]
      try:  # remove duplicate site, otherwise it will do nasty things where processing the request
        failovers.remove(targetSE)
      except ValueError:
        pass
      random.shuffle(failovers)
      metadata['resolvedSE'] = failovers
      resultFailover = failoverTransfer.transferAndRegisterFileFailover(fileName=fileName,
                                                                        localPath=metadata['localpath'],
                                                                        lfn=metadata['lfn'],
                                                                        targetSE=targetSE,
                                                                        failoverSEList=metadata['resolvedSE'],
                                                                        fileMetaDict=metadata['filedict'],
                                                                        fileCatalog=catalogs)
      if not resultFailover['OK']:
        LOG.error('Could not transfer and register %s with metadata:\n %s' %
                       (fileName, pformat(metadata['filedict'])))
        cleanUp = True
        break  # no point continuing if one completely fails

    os.remove("DISABLE_WATCHDOG_CPU_WALLCLOCK_CHECK")  # cleanup the mess

    self.workflow_commons['Request'] = failoverTransfer.request

    # If some or all of the files failed to be saved to failover
    if cleanUp:
      lfns = []
      for fileName, metadata in final.items():
        lfns.append(metadata['lfn'])

      result = self._cleanUp(lfns)
      return S_ERROR('Failed to upload output data')

    return S_OK('Output data uploaded')

  def _getExtensionDictionary(self):
    """Return dictionary of extensions and patterns for the respective files."""
    operationsValue = self.ops.getOptionsDict('/Production/ExtensionToFiletypes')
    if not operationsValue['OK']:
      # the default value, for ilc VO
      return {'.stdhep': ['_gen'], '.slcio': ['_sim', '_rec', '_dst']}
    opValue = operationsValue['Value']
    resultDict = {}
    for extension, patterns in opValue.items():
      resultDict[extension] = [pattern.strip() for pattern in patterns.split(',')]
    return resultDict

  def _expectedExtension(self, filename):
    """Return the expected extension based on the production type hinted in the filename."""
    extensionTofiletype = self._getExtensionDictionary()
    for extension, patterns in extensionTofiletype.items():
      if any(pattern in filename for pattern in patterns):
        LOG.info("Matched %s with %s: expecting %s" % (patterns, filename, extension))
        return extension
    LOG.warn("Unknown production file type: %s" % filename)
    return ''

  def getTreatedOutputlistNew(self, producedData, treatedOutputlist, outputfileObject):
    """Return properly formated outputList in treatedOutputlist.

    TODO: describe what a properly formatted outputList looks like so we know what this function actually does.
    """
    expectedOutputFile, dummy_ext = self.getBasenameAndExtension(outputfileObject['outputFile'].lower())
    for productionFile in producedData:
      LOG.debug("Prodfile %s; outFile %s" % (productionFile, expectedOutputFile))
      productionFile, extension = self.getBasenameAndExtension(productionFile)
      LOG.debug("Removed extension: %s" % productionFile)
      if productionFile in treatedOutputlist:
        # This has already been treated, no need to come back to it.
        continue
      appdict = {}
      extensionTofiletype = self._getExtensionDictionary()
      fileTypes = []
      for _extension, patterns in extensionTofiletype.items():
        fileTypes.extend(patterns)
      for fType in fileTypes:
        # No idea why the second thing is necessary, but it is there in the original function
        if fType in expectedOutputFile and (fType != '_dst' or '_dst_' in productionFile.lower()):
          filePrototype = outputfileObject['outputFile'].split(fType)[0]
          if filePrototype in productionFile and fType in productionFile:
            appdict.update(outputfileObject)
            appdict['outputFile'] = productionFile + extension
            treatedOutputlist[productionFile] = appdict
            if fType in ('_rec', '_dst'):  # there will only be one _rec or _dst file...
              return

  def getBasenameAndExtension(self, filepath):
    """returns tuple of basename and extenion."""
    baseFileName = os.path.basename(filepath)
    extension = self._expectedExtension(baseFileName)
    baseFileNameWoExtension = baseFileName.replace(extension, "")
    return baseFileNameWoExtension, extension

#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Job Reset Agent takes care of the following cases.

* Finds jobs stuck in staging status and reschedules them if associated files are already staged
* Finds production jobs in completed status:

  * Marks a Job Done if request status is Done
  * Marks a Job Done if no request is found
  * Marks a Job Done if ReplicateAndRegister operation has an error "No such file"
  * Marks a Job DOne if RemoveReplica cannot be completed. Let DataRecoveryAgent take over
  * Resets Failed requests with ReplicateAndRegister operations if error is other than "No such file"

* Finds production jobs in failed status with minor status "Pending Requests"

  * Marks a Job Failed with minor status "Requests Done" if request status is Done
  * Marks a Job Failed with minor status "Requests Done" if no request is found
  * Resets Failed RemoveFile requests

* Finds user jobs in completed status:

  * Marks a Job Done if there is no request and minorStatus and appStatus are in final states
  * Marks a Job Done if request status is Done
  * Resets requests if the request status is other than Done
  * Unregister files if a new file is supposed to be registered
"""

from __future__ import absolute_import
from collections import defaultdict
from datetime import datetime, timedelta
from pprint import pformat

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities.PrettyPrint import printTable
from DIRAC.Core.Utilities.Proxy import UserProxy
from DIRAC.Core.Base.Client import Client

from DIRAC.FrameworkSystem.Client.NotificationClient import NotificationClient
from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.WorkloadManagementSystem.Client.JobMonitoringClient import JobMonitoringClient
from DIRAC.WorkloadManagementSystem.Client.JobManagerClient import JobManagerClient
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.Resources.Catalog.FileCatalogFactory import FileCatalogFactory

from ILCDIRAC.Core.Utilities.LFNPathUtilities import cleanUpLFNPath
from six.moves import map

__RCSID__ = "$Id$"

AGENT_NAME = 'WorkloadManagement/JobResetAgent'

FINAL_APP_STATES = ["Job Finished Successfully",
                    "Unknown"]

FINAL_MINOR_STATES = ["Pending Requests",
                      "Application Finished Successfully"]

NO_RESET = 'NO_RESET'


class JobResetAgent(AgentModule):
  """JobResetAgent."""

  def __init__(self, *args, **kwargs):
    AgentModule.__init__(self, *args, **kwargs)
    self.name = 'JobResetAgent'
    self.enabled = True

    self.userJobTypes = ['User']
    self.prodJobTypes = ['MCGeneration', 'MCSimulation', 'MCReconstruction', 'MCReconstruction_Overlay', 'Split',
                         'MCSimulation_ILD', 'MCReconstruction_ILD', 'MCReconstruction_Overlay_ILD', 'Split_ILD']

    self.addressTo = ["ilcdirac-admin@cern.ch"]
    self.addressFrom = "ilcdirac-admin@cern.ch"
    self.emailSubject = "JobResetAgent"

    self.accounting = defaultdict(list)
    self.errors = []

    self.nClient = NotificationClient()
    self.reqClient = ReqClient()
    self.jobMonClient = JobMonitoringClient()
    self.dataManager = DataManager()
    self._jobDB = None

    self.jobStateUpdateClient = Client(useCertificates=True, timeout=10)
    self.jobStateUpdateClient.setServer('WorkloadManagement/JobStateUpdate')
    self.jobManagerClient = JobManagerClient(useCertificates=True, timeout=10)

    self._fcClient = None

  @property
  def jobDB(self):
    """Create JobDB on demand.

    Avoid importing JobDB at importing of this agent, which gives some errors for the documentation
    """
    if not self._jobDB:
      from DIRAC.WorkloadManagementSystem.DB.JobDB import JobDB
      self._jobDB = JobDB()
    return self._jobDB

  @property
  def fcClient(self):
    """Create a FileCatalogClient on demand."""
    if not self._fcClient:
      fcType = "FileCatalog"
      result = FileCatalogFactory().createCatalog(fcType)
      if not result['OK']:
        raise RuntimeError("Cannot instantiate FileCatalog")
      self._fcClient = result['Value']

    return self._fcClient

  def beginExecution(self):
    """Reload the configurations before every cycle."""

    self.enabled = self.am_getOption('EnableFlag', False)
    self.addressTo = self.am_getOption('MailTo', self.addressTo)
    self.addressFrom = self.am_getOption('MailFrom', self.addressFrom)
    self.userJobTypes = self.am_getOption('UserJobs', self.userJobTypes)
    self.prodJobTypes = self.am_getOption('ProdJobs', self.prodJobTypes)
    self.accounting.clear()

    return S_OK()

  def sendNotification(self):
    """sends email notification about job treatments."""

    if not(self.errors or self.accounting):
      return S_OK()

    emailBody = ""
    rows = []
    for jobType, val in self.accounting.items():
      emailBody += "\nTotal number of %s Jobs treated: %s\n\n" % (jobType, len(val))
      for job in val:
        rows.append([[str(job['JobID'])], [job['JobStatus']], [job['Treatment']]])

      if rows:
        columns = ["Job ID", "Job Status", "Treatment"]
        emailBody += printTable(columns, rows, printOut=False, numbering=False, columnSeparator=' | ')
        rows = []

    if self.errors:
      emailBody += "\n\nErrors:"
      emailBody += "\n".join(self.errors)

    self.log.notice(emailBody)
    for address in self.addressTo:
      res = self.nClient.sendMail(address, self.emailSubject, emailBody, self.addressFrom, localAttempt=False)
      if not res['OK']:
        self.log.error("Failure to send Email notification to ", address)
        continue

    self.errors = []
    self.accounting.clear()
    return S_OK()

  def logError(self, errStr, varMsg=''):
    self.log.error(errStr, varMsg)
    self.errors.append(errStr + varMsg)

  def getJobs(self, status, jobType=None, minorStatus=None):
    """returns jobs with a given status, job type, minor status and lastUpdateTime older than 1 day."""

    attrDict = dict(Status=status)
    if jobType:
      attrDict['JobType'] = jobType

    if minorStatus:
      attrDict['MinorStatus'] = minorStatus

    time = datetime.now() - timedelta(days=1)
    res = self.jobDB.selectJobs(attrDict, older=time)
    if not res['OK']:
      self.logError("Failure to get Jobs", res['Message'])
      return res

    jobIDs = list(map(int, res['Value']))
    return S_OK(jobIDs)

  def treatUserJobWithNoReq(self, jobID):
    """treatment for user jobs in completed status which don't have a request in RMS."""

    self.log.notice("No request found for job: %s" % jobID)
    res = self.jobMonClient.getJobsMinorStatus([jobID])
    if not res['OK']:
      self.logError("Failure to get Minor Status", "Job ID: %s, Message: %s" % (jobID, res['Message']))
      return res

    minorStatus = res['Value'][jobID]['MinorStatus']

    res = self.jobMonClient.getJobsApplicationStatus([jobID])
    if not res['OK']:
      self.logError("Failure to get Application Status", "Job ID: %s, Message: %s" % (jobID, res['Message']))
      return res

    appStatus = res['Value'][jobID]['ApplicationStatus']

    if minorStatus in FINAL_MINOR_STATES and appStatus in FINAL_APP_STATES:
      res = self.markJob(jobID, "Done")
      if res["OK"]:
        self.accounting["User"].append({"JobID": jobID, "JobStatus": "Completed", "Treatment": ("Job Marked Done "
                                        "because Minor Status: '%s' and Application Status: '%s' are in Final States" %
                                                                                                (minorStatus, appStatus))})
      return res

    self.logError("Something not as expected for Job Status, ", "please check: %s" % jobID)
    return S_OK()

  def treatUserJobWithReq(self, jobID, request):
    """treatment for user jobs in completed status which have a request in RMS."""

    if request.Status == "Done":
      self.log.notice("Request is Done: %s " % request)
      res = self.markJob(jobID, "Done")
      if res["OK"]:
        self.accounting["User"].append({"JobID": jobID, "JobStatus": "Completed", "Treatment": ("Job Marked Done "
                                        "because associated Request with ID: %s is Done" % request.RequestID)})
      return res

    if request.Status in ("Waiting", "Scheduled"):
      self.log.notice("Request is Waiting (for FTS): %s " % request)
      return S_OK()

    self.log.notice("Request not Done: %s " % request)

    if request.Status in ("Failed",):
      res = self._treatFailedUserRequests(jobID, request)
    if not res['OK']:
      return res

    if res['Value'] is None or res['Value'] != NO_RESET:
      res = self.resetRequest(request.RequestID)
      if res['OK']:
        self.accounting['User'].append({'JobID': jobID, 'JobStatus': 'Completed', 'Treatment': ('Resetting request '
                                        'with ID: %s and Status: %s' % (request.RequestID, request.Status))})
    return res

  def treatFailedProdWithReq(self, jobID, request):
    """treatment for production jobs in failed status which have a request in RMS."""
    if request.Status == 'Done':
      self.log.notice('Request is Done: %s ' % request)
      res = self.markJob(jobID, "Failed")
      if res["OK"]:
        self.accounting["Production"].append({"JobID": jobID, "JobStatus": "Failed with 'Pending Requests'",
                                              "Treatment": ("Job Marked Failed because associated Request "
                                                            "with ID: %s is Done" % request.RequestID)})
      return res

    for op in request:
      lfns = "\n\t".join(lfn.LFN for lfn in op)
      self.log.notice('Operation for failed job: %s, %s, %s, %s\n\t%s' %
                      (request.RequestID, op.Type, op.Status, op.Error, lfns))

      if op.Type == "RemoveFile" and op.Status == 'Failed':
        filesToRemove = [lfn.LFN for lfn in op]
        self.log.notice("Removing files %s" % filesToRemove)
        if self.enabled:
          res = self.dataManager.removeFile(filesToRemove, force=True)
          if not res["OK"]:
            self.logError("Failure to remove Files", ":%s Message: %s" % (filesToRemove, res["Message"]))

          res = self.resetRequest(request.RequestID)
          if res["OK"]:
            self.accounting["Production"].append({"JobID": jobID, "JobStatus": "Failed with 'Pending Requests'",
                                                  "Treatment": ("Resetting request with ID: %s and Status: %s"
                                                                % (request.RequestID, request.Status))})
          return res
      elif op.Status == "Failed":
        self.log.notice("Can't handle operation of type: %s" % op.Type)

    return S_OK()

  def treatFailedProdWithNoReq(self, jobID):
    """treatment for production jobs in failed status which don't have a request in RMS."""

    res = self.markJob(jobID, "Failed")
    if res["OK"]:
      self.accounting["Production"].append({"JobID": jobID,
                                            "JobStatus": "Failed with 'Pending Requests'",
                                            "Treatment": ("Job Marked Failed with minor status 'Requests Done' "
                                                          "and application status 'CompletedJobChecker' because no associated "
                                                          "request is found")})
    return res

  def treatCompletedProdWithReq(self, jobID, request):
    """treatment for production jobs in completed status which have a request in RMS."""

    if request.Status == "Done":
      self.log.notice("Request is Done: %s " % request)
      res = self.markJob(jobID, "Done")
      if res["OK"]:
        self.accounting["Production"].append({"JobID": jobID, "JobStatus": "Completed", "Treatment": (
            "Job Marked " "Done because associated Request with ID: %s is Done" % request.RequestID)})
      return res

    if request.Status in ("Waiting", "Scheduled"):
      self.log.notice("Request is Waiting (for FTS): %s " % request)
      return S_OK()

    for op in request:
      self.log.info("Operation for completed job: %s, %s, %s, %s" %
                    (request.RequestID, op.Type, op.Status, op.Error))
      if op.Type == 'ReplicateAndRegister' and op.Status == 'Failed':
        # Check if it failed because the file(s) no longer exists
        lfnStatuses = [lfn.Error is not None and "No such file" in lfn.Error for lfn in op]
        lfnErrors = [(lfn.LFN, lfn.Error) for lfn in op]
        self.log.notice("LFN Status: %s" % lfnErrors)
        if all(lfnStatuses):
          res = self.markJob(jobID, "Done")
          if res["OK"]:
            self.accounting["Production"].append({"JobID": jobID,
                                                  "JobStatus": "Completed",
                                                  "Treatment": "Job marked Done because no file is found for operation",
                                                  })
          return res
        res = self.resetRequest(request.RequestID)
        if res["OK"]:
          self.accounting["Production"].append({"JobID": jobID, "JobStatus": "Completed", "Treatment": ("Resetting "
                                                "request with ID: %s and Status: %s" % (request.RequestID,
                                                                                        request.Status))})
        return res

      if op.Type == 'RemoveReplica' and op.Status == 'Failed':
        # Just mark the job as Done, because we cannot remove a single replica,
        # let DataRecoveryAgent take care of the rest
        lfnErrors = [(lfn.LFN, lfn.Error) for lfn in op]
        self.log.notice("RemoveReplica LFN Status: %s" % lfnErrors)
        res = self.markJob(jobID, "Done")
        if res["OK"]:
          self.accounting["Production"].append({"JobID": jobID,
                                                "JobStatus": "Completed",
                                                "Treatment": "Job marked Done because RemoveReplica failed",
                                                })
          return res

      elif op.Status == "Failed":
        self.log.notice("Cannot handle Operation Type: %s" % op.Type)

    return S_OK()

  def treatCompletedProdWithNoReq(self, jobID):
    """marks the job done."""
    res = self.markJob(jobID, "Done")
    if res["OK"]:
      self.accounting["Completed"].append({"JobID": jobID, "JobStatus": "Completed", "Treatment": ("Job Marked "
                                           "Done because no associated request is found")})
    return res

  def _treatFailedUserRequests(self, jobID, request):
    """Treat failed user Requests."""
    for op in request:
      self.log.info('Operation for failed job: %s, %s, %s, %s\n\t%s' %
                    (request.RequestID, op.Type, op.Status, op.Error,
                     "\n\t".join(lfn.LFN for lfn in op)))
      failed = False
      if op.Type not in ('RegisterFile', 'ReplicateAndRegister') or op.Status != 'Failed':
        return S_OK()
      if op.Type == 'RegisterFile':
        for lfn in op:
          if lfn.Error is None or "File already registered with alternative metadata" not in lfn.Error:
            continue
          self.log.info("Getting user proxy %s (%s)" % (request.OwnerDN, request.OwnerGroup))
          with UserProxy(proxyUserGroup=request.OwnerGroup, proxyUserDN=request.OwnerDN) as res:
            if res['OK']:
              self.log.info('Unregistering LFN: %s' % lfn.LFN)
              res = self.fcClient.removeFile(lfn.LFN)
          self.log.info("Tried unregistering:\n%s" % pformat(res))
          if not res['OK']:
            self.logError("Failed to unregister file", "%s: %s" % (lfn.LFN, res['Message']))
            failed = True
            continue
          if res['Value'].get('Failed', {}):
            for failedLFN, reason in res['Value']['Failed'].items():
              self.logError("Failed to unregister file", "%s: %s" % (failedLFN, reason))
            failed = True
            continue
          self.accounting["User"].append({"JobID": jobID,
                                          "JobStatus": "Completed",
                                          "Treatment": "Unregistered LFN %s" % lfn.LFN})

        if failed:
          return S_ERROR("Failed to unregister some files")

      elif op.Type == 'ReplicateAndRegister':
        for lfn in op:
          # if any file is missing, we fail the job
          if lfn.Error is None or 'No such file or directory' not in lfn.Error:
            continue
          res = self.markJob(jobID, 'Failed')
          if res["OK"]:
            self.accounting["User"].append({"JobID": jobID, "JobStatus": "Completed",
                                            "Treatment": "Job Marked Failed because Request cannot be finished, file not found"})
            return S_OK(NO_RESET)

      return S_OK()

  def checkJobs(self, jobIDs, treatJobWithNoReq, treatJobWithReq):
    """executes treatment functions for jobs with and without requests."""

    res = self.reqClient.readRequestsForJobs(jobIDs)
    if not res['OK']:
      self.logError('Failure to read requests for jobs', res['Message'])
      return res

    result = res['Value']
    for jobID in jobIDs:
      if ((jobID not in result['Successful'] and jobID not in result['Failed']) or (
              jobID in result['Failed'] and 'Request not found' in result['Failed'][jobID])):
        self.log.notice("No request found for job: %s" % jobID)
        treatJobWithNoReq(jobID)

      elif jobID in result['Successful']:
        self.log.notice("Found the request for Job: %s " % jobID)
        request = result['Successful'][jobID]
        treatJobWithReq(jobID, request)

    return S_OK()

  def getStagedFiles(self, lfns):
    """returns a list of staged files."""
    if not lfns:
      self.log.notice("No LFNs passed to check staging status")
      return S_OK()

    voName = lfns[0].split('/')[1]
    se = StorageElement("CERN-SRM", vo=voName)
    res = se.getFileMetadata(lfns)
    if not res["OK"]:
      self.logError("Failure to getFileMetadata for LFNs", "%s" % lfns)
      return res

    stagedFiles = [lfn for lfn, val in res["Value"]["Successful"].items() if val["Cached"] > 0]
    return S_OK(stagedFiles)

  def getInputDataForJobs(self, jobList):
    """returns the input data for a given list of jobIDs."""
    inputData = defaultdict(list)
    for jobID in jobList:
      res = self.jobMonClient.getInputData(jobID)
      if not res['OK']:
        self.logError("Failure to get input data for", "JobID: %s, Message: %s" % (jobID, res["Message"]))
        continue

      for lfn in res['Value']:
        lfn = cleanUpLFNPath(lfn)
        inputData[lfn].append(jobID)

    return S_OK(inputData)

  def rescheduleJobs(self, jobsToReschedule):
    """resets a list of jobs."""
    result = dict(Failed=[], Successful=[])
    for job in jobsToReschedule:
      res = self.jobManagerClient.resetJob(job)
      if res['OK']:
        result['Successful'].append(job)
      else:
        self.logError("Failed to reset job", "%s: %s" % (job, res['Message']))
        result['Failed'].append(job)

    self.log.info("Reset jobs: %s" % result)
    return S_OK(result)

  def checkStagingJobs(self, jobList):
    """gets input data and stager status, then reschedules jobs whose associated files are already staged."""

    res = self.getInputDataForJobs(jobList)
    inputData = res['Value']

    if not inputData:
      self.log.notice("No input data found for job list %s" % jobList)
      return S_OK()

    self.log.notice("Input Data found: %s" % inputData)
    res = self.getStagedFiles(list(inputData.keys()))
    if not res['OK']:
      return res

    stagedFiles = res['Value']

    jobsToReschedule = set()
    for lfn in stagedFiles:
      jobsToReschedule.add(inputData[lfn])
      self.log.notice("Jobs to be rescheduled: %s" % jobsToReschedule)

      if self.enabled and jobsToReschedule:
        res = self.rescheduleJobs(jobsToReschedule)
        if res["OK"]:
          for jobID in res["Value"]["Successful"]:
            self.accounting["Staging"].append({"JobID": jobID, "JobStatus": "Staging", "Treatment": (
                                               "Job Rescheduled because associated files are already Staged")})

    return S_OK()

  def resetRequest(self, requestID):
    """resets failed requests for a given requestID."""
    if not self.enabled:
      return S_OK()

    res = self.reqClient.resetFailedRequest(requestID, allR=True)
    if not res["OK"]:
      self.logError("Failed to reset request", "Request ID: %s, Message: %s" % (requestID, res['Message']))
      return res

    if res["Value"] == "Not reset":
      self.logError("Failed to reset request", "Request ID: %s" % (requestID))
      return S_ERROR()

    self.log.notice("Request %s is successfully reset" % requestID)
    return S_OK()

  def markJob(self, jobID, status, minorStatus="Requests Done", application="CompletedJobChecker"):
    """marks a job with given status, minorStatus and application."""

    self.log.notice("Marking job %s as %s" % (jobID, status))

    if not self.enabled:
      return S_OK()

    res = self.jobStateUpdateClient.setJobStatus(jobID, status, minorStatus, application)
    if not res["OK"]:
      self.logError("Failed to mark ", "Job: %s as %s, Error: %s" % (jobID, status, res['Message']))
      return res

    self.log.notice("Job %s is successfully maked as %s" % (jobID, status))
    return S_OK()

  def execute(self):
    """main execution loop of Agent."""

    # process completed prod jobs
    res = self.getJobs(status="Completed", jobType=self.prodJobTypes)
    if res["OK"]:
      completedJobIDs = res["Value"]
      if completedJobIDs:
        self.checkJobs(jobIDs=completedJobIDs,
                       treatJobWithNoReq=self.treatCompletedProdWithNoReq,
                       treatJobWithReq=self.treatCompletedProdWithReq)
      else:
        self.log.notice("No production jobs found with Completed status")

    # process failed prod jobs
    res = self.getJobs(status="Failed", jobType=self.prodJobTypes, minorStatus="Pending Requests")
    if res["OK"]:
      failedJobIDs = res["Value"]
      if failedJobIDs:
        self.checkJobs(jobIDs=failedJobIDs,
                       treatJobWithNoReq=self.treatFailedProdWithNoReq,
                       treatJobWithReq=self.treatFailedProdWithReq)
      else:
        self.log.notice("No production jobs found with Failed status and pending requests")

    # process completed user jobs
    res = self.getJobs(status="Completed", jobType=self.userJobTypes)
    if res["OK"]:
      completedUserJobs = res["Value"]
      if completedUserJobs:
        self.checkJobs(jobIDs=completedUserJobs,
                       treatJobWithNoReq=self.treatUserJobWithNoReq,
                       treatJobWithReq=self.treatUserJobWithReq)
      else:
        self.log.notice("No user jobs found with Completed status")

    # process STAGING jobs
    res = self.getJobs(status="Staging")
    if res["OK"]:
      stagingJobs = res["Value"]
      if stagingJobs:
        self.checkStagingJobs(stagingJobs)
      else:
        self.log.notice("No staging jobs found")

    # send email notification
    self.sendNotification()

    return S_OK()
